/* $Id: posix_memalign.c 2741 2022-09-17 21:58:12Z teshields $ */

/* This file is part of the Libnew package that provides replacement
   implementations for core standard C library memory management
   functions, and for the standard C++ 'new()' and 'delete()' operators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#if defined (TRACE)
#include <stdio.h>
#endif

#include <stdlib.h>

#if defined (USE_BDWGC)
#include <gc/gc.h>
#elif defined (USE_DMALLOC)
#define DMALLOC_DISABLE
#include <errno.h>
#include <dmalloc.h>
#else
#error "Neither USE_BDWGC nor USE_DMALLOC is defined"
#endif

int posix_memalign (void** memptr, size_t alignment, size_t size)
{
#if defined (TRACE)
#if defined (USE_BDWGC)
  int result = GC_posix_memalign (memptr, alignment, size);
#elif defined (USE_DMALLOC)
  *memptr = dmalloc_malloc (NULL, 0,
			    size,
			    DMALLOC_FUNC_MEMALIGN,
			    alignment,
			    0 /* no xalloc messages */
			   );
  if (errno == EINVAL || errno == ENOMEM)
     result = errno;
  else
     result = 0;
#endif
  fprintf (stdout, "posix_memalign (%#lx [@%#lx], %lu, %lu)=(%#lx [@%#lx], %i)\n",
		   *memptr, memptr, alignment, size, *memptr, memptr, result);
  return result;
#else
#if defined (USE_BDWGC)
  return GC_posix_memalign (memptr, alignment, size);
#elif defined (USE_DMALLOC)
  *memptr = dmalloc_malloc (NULL, 0,
			    size,
			    DMALLOC_FUNC_MEMALIGN,
			    alignment,
			    0 /* no xalloc messages */
			   );
  if (errno == EINVAL || errno == ENOMEM)
     return errno;
  else
     return 0;
#endif
#endif
}
