/* $Id: valloc.c 2740 2022-09-17 21:57:37Z teshields $ */

/* This file is part of the Libnew package that provides replacement
   implementations for core standard C library memory management
   functions, and for the standard C++ 'new()' and 'delete()' operators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#if defined (TRACE)
#include <stdio.h>
#endif

#include <stdlib.h>

#if defined (USE_BDWGC)
#include <gc/gc.h>
#elif defined (USE_DMALLOC)
#define DMALLOC_DISABLE
#include <unistd.h>
#include <dmalloc.h>
#warning "The DMALLOC package provides an implementation of 'valloc()'."
#else
#error "Neither USE_BDWGC nor USE_DMALLOC is defined"
#endif

void* valloc (size_t size)
{
#if defined (TRACE)
#if defined (USE_BDWGC)
  void* result = GC_MALLOC (size);
#elif defined (USE_DMALLOC)
  static int pagesize = 0;
  if (pagesize == 0) pagesize = getpagesize ();
  void* result = dmalloc_malloc (NULL, 0,
				 size,
				 DMALLOC_FUNC_VALLOC,
				 size /* alignment */,
				 0 /* no xalloc messages */
				);
#endif
  fprintf (stdout, "malloc (%lu)=%#lx\n", size, result);
  return result;
#else
#if defined (USE_BDWGC)
  return GC_MALLOC (size);
#elif defined (USE_DMALLOC)
  return dmalloc_malloc (NULL, 0,
			 size,
			 DMALLOC_FUNC_VALLOC,
			 size /* alignment */,
			 0 /* no xalloc messages */
			);
#endif
#endif
}
