%language "C++"
%require "3.8.2"
%skeleton "glr2.cc"

%glr-parser

%define lr.type ielr
%define lr.default-reduction consistent

%define api.token.raw

//%define parse.assert
%define parse.trace
%define parse.error custom

%locations

%initial-action
{
  @$.initialize (&theFiles.inName);

  yoxOutGdefs.setCurPos (@$.begin);
  yoxOutG.setCurPos (@$.begin);
  yoxOutGimport.setCurPos (@$.begin);

  yoxOutH.setCurPos (@$.begin);
  yoxOutHrequires.setCurPos (@$.begin);
  yoxOutHprovides.setCurPos (@$.begin);

  actionCLOSER.setCurPos (@$.begin);
  attrDECL.setCurPos (@$.begin);
  possibleLHS.setCurPos (@$.begin);
  ruleDIRECTIVE.setCurPos (@$.begin);
  unionBODY.setCurPos (@$.begin);
  yaccCODE.setCurPos (@$.begin);

  separatorMARK.setCurPos (@$.begin);
}

%header
%define api.namespace {AGC}
%code requires {#include "location.hh"}
%define api.location.type {AGC::location}

%parse-param {AGC::bisonLexer& theLexer}
%parse-param {yoxControl& ctrl}
%parse-param {specIO& theFiles}

%expect 1

%verbose

%{
/* $Id: agc-bison.yy 2842 2022-12-13 01:19:44Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
%}

%code top
{
// hack to avoid a 'yydebug' global in every parser
#define yydebug agc_bison_debug
}

%code requires
{
#include "lextypes.hh"
// attrbTriple:     lexical representation of an attribute occurrence
// numberRep:       lexical representation of a number

#include "semtypes.hh"
// attrbPair:       semantic representation of an attribute occurence
// attrbRefSeqElem: attribute occurrence w/location
// attrbRefSeq:     sequence of occurrences in an attribute definition

#include "scan.hh"

namespace AGC
{
class bisonLexer;
}
#include "agc-Yfile.hh"
}

%code provides
{
#if defined(WITH_REFLEX)
#include <reflex/input.h>
#endif
// called from 'main ()'
extern "C" void
agc_bison_parse (
#if defined(WITH_REFLEX)
                          const reflex::Input&,
#elif defined(WITH_FLEX)
                          std::ifstream&,
#else
#error "missing -DWITH_FLEX or -DWITH_REFLEX compile option"
// assume -DWITH_FLEX to avoid cascading errors
                          std::ifstream&,
#endif
                          Yfile_specIO&,
                          int);
}

%define api.value.type variant



// Ox tokens: common to lexer & parser specs

%token                OX_LBRACE      "@{"
%token                OX_RBRACE      "@}"
                                   // @<grammar symbol>(.<n>)?.<attribute name>@
%token <attrbTriple>  OX_ATTRB_REF   "attribute reference"
                                   // grammar symbol name, C/C++ identifier,
                                   // Ox macro name, Ox attribute name
%token <name_literal> ID             "identifier"

%token <char_literal> LITERAL_CHAR   "character literal"
%token <char_literal> LITERAL_STRING "string literal"



// non-alpha character tokens common to lexer & parser specs

%token                LBRACE         "{"
%token                RBRACE         "}"
%token                SEMICOLON      ";"



// Lex/Yacc & Lex/Yacc-extension reserved tokens

%token                MARK           "%%"
%token                LCURL          "%{"
%token                RCURL          "%}"



// Ox tokens: lexer specs

%token <name_literal> TERM_ID        "terminal symbol"
%token <char_literal> TERM_LITERAL   "terminal literal"



// lexer: C/C++ keyword tokens

%token                RETURN         "return"



// lexer: miscellaneous tokens

%token                NEWLINE        "\\n"
%token                SOSF           "start of generic scanner source file"
%token                NO_ATTRBREFSEC "no attribute reference section"



// Ox tokens: parser specs
                                    // <grammar symbol>(.<n>)?.<attribute name>
%token <attrbTriple>  OX_ATTRB_OCCUR "attribute occurrence"
%token <keyword>      OX_ATTRIBUTES  "@attributes"
%token <keyword>      OX_MACRO       "@macro"
%token                OX_END         "@end"
%token <keyword>      OX_TRAV        "@traversal"
%token                OX_PREORDER    "@preorder"
%token                OX_POSTORDER   "@postorder"
%token                OX_LEFTTORIGHT "@lefttoright"
%token                OX_RIGHTTOLEFT "@righttoleft"
%token                OX_REVORDER    "@revorder"
%token                OX_REVDIR      "@revdirection"
%token                OX_DISABLE     "@disable"
%token <keyword>      OX_AUTOSYN     "@autosyn"
%token <keyword>      OX_AUTOINH     "@autoinh"
%token <keyword>      OX_WARN        "@warn"
%token                OX_I           "@i"
%token                OX_M           "@m"
%token                OX_E           "@e"
%token                OX_DECORATE    "@decorate"
%token                OX_VOID        "@void"
                                   // <traversal name>
%token <name_literal> OX_ID          "traversal mode annunciator"



// parser: non-alpha character tokens

%token                COMMA          ","
%token                LPAREN         "("
%token                RPAREN         ")"
%token                LBRACKET       "["
%token                RBRACKET       "]"
                                   // MSTA named reference
%token                AT             "@"
%token                ELLIPSIS       "..."
                                   // MSTA charset specifications
%token                LTARROW        "<-"
%token                DBLARROW       "<->"
                                   // Bison predicate action
%token                PREDICATE      "%?"
                                   // grammar symbol, (dotted) name
%token <name_literal> GS_ID          "nonterminal symbol"
%token <name_literal> NAMESPACE_ID   "C++ qualified name"
%token <char_literal> LITERAL_PATH   "literal path name"
                                   // $ID or $NUM (BtYacc, BYacc)
%token <name_literal> INHATTR_ID     "inherited attribute"
%token <name_literal> API_VARIABLE   "bison API variable"
%token <name_literal> API_VALUE      "bison API value"
%token <numberRep>    NUM            "number" // [0-9]+



// parser: Yacc & Yacc-extension reserved words

%token <keyword>      LEFT           "%left"
%token <keyword>      RIGHT          "%right"
%token <keyword>      NONASSOC       "%nonassoc"
%token <keyword>      TOKEN          "%token"
%token                PREC           "%prec"
%token <keyword>      TYPE           "%type"
%token <keyword>      NTERM          "%nterm" // Bison
%token <keyword>      START          "%start"
%token <keyword>      UNION          "%union"
%token                EMPTY          "%empty" // Bison



// parser: Yacc-extension directives

%token <keyword>      EXPECT         "%expect"      // Bison, BYacc, MSTA
%token <keyword>      EXPECT_RR      "%expect-rr"   // Bison, BYacc
%token <keyword>      PURE_PARSER    "%pure-parser" // Bison, BYacc
%token                DPREC          "%dprec"       // Bison GLR parsing
%token                MERGE          "%merge"       // Bison GLR parsing
%token <keyword>      PRECEDENCE     "%precedence"  // Bison
%token                LOOKAHEAD      "%la"          // MSTA
%token <keyword>      SCANNER        "%scanner"     // MSTA (not supported)



// parser: Yacc-extension composite directives

%token <keyword>      CODE         // %code, %initial-action   // Bison, BYacc
%token <keyword>      LOCATION     // %location, %position, %@ // BtYacc
%token <keyword>      DESTRUCTOR     "%destructor"             // Bison, BYacc
%token <keyword>      LEX_PARAM      "%lex-param"              // Bison, BYacc
%token <keyword>      PARSE_PARAM    "%parse-param"            // Bison, BYacc
%token <keyword>      DEFINE         "%define"                 // Bison
%token <keyword>      LANGUAGE       "%language"               // Bison
%token <keyword>      NAME_PREFIX    "%name-prefix"            // Bison
%token <keyword>      PARAM          "%param"                  // Bison
%token <keyword>      PRINTER        "%printer"                // Bison
%token <keyword>      SKELETON       "%skeleton"               // Bison
%token <keyword>      LOCAL          "%local"                  // MSTA
%token <keyword>      EXPORT         "%export"                 // MSTA
%token <keyword>      IMPORT         "%import"                 // MSTA



// parser: C/C++ keyword tokens

%token <keyword>      CONST          "const"
%token <keyword>      VOLATILE       "volatile"
%token <keyword>      RESTRICT       "restrict" // C99
%token <keyword>      REGISTER       "register"
%token <keyword>      STATIC         "static"
%token <keyword>      TYPE_KEYWORD            // C/C++ type keywords
%token <keyword>      ELABORATED_TYPE_KEYWORD // C/C++ composite type keywords



// parser: C/C++ operator tokens

%token                PERIOD         "."
%token                RTARROW        "->" // & MSTA charset specifications
%token                PLUSPLUS       "++"
%token                MINUSMINUS     "--"
%token                TILDE          "~"
%token                EXCLAMATION    "!"
%token                MINUS          "-"
%token                PLUS           "+"
%token                AMPERSAND      "&" // & C/C++ type specification
%token                ASTERISK       "*" // & C/C++ type specification
%token                SLASH          "/"
%token                PERCENT        "%"
%token                LTSHIFT        "<<"
%token                RTSHIFT        ">>"
%token                LESSTHAN       "<" // & Yacc type tag, Bison '%merge',
                                         //   C++ template types
%token                GTRTHAN        ">" // & Yacc type tag, Bison '%merge',
                                         //   C++ template types
%token                LTOREQ         "<="
%token                GTOREQ         ">="
%token                EQUAL          "=="
%token                NOTEQUAL       "!="
%token                UPARROW        "^"
%token                VBAR           "|"
%token                LOGAND         "&&"
%token                LOGOR          "||"
%token                QUESTION       "?"
%token                COLON          ":" // & C/C++ type specification
%token                ASSIGN         "="
%token                PLUSEQ         "+="
%token                MINUSEQ        "-="
%token                TIMESEQ        "*="
%token                DIVEQ          "/="
%token                REMEQ          "%="
%token                LTSHIFTEQ      "<<="
%token                RTSHIFTEQ      ">>="
%token                BITANDEQ       "&="
%token                BITXOREQ       "^="
%token                BITOREQ        "|="



// parser: C++ operator tokens

%token                PTRTOMEMBER    "->*"
%token                MEMBERSELECT   ".*"
%token                SCOPE          "::"



// parser: miscellaneous tokens

%token                END 0          "end of file"



%printer
{
  yyo << "\"" + $$ + "\"";
} <name_literal>
%printer
{
  yyo << "\"" + $$ + "\"";
} <char_literal>
%printer
{
  yyo << "\"" + $$ + "\"";
} <keyword>

%start                AG_spec

%type <name_literal>  terminal_symbol

%type <attrbPair>     attrb_occur attrb_ref
%type <attrbRefSeq>   attrb_evaluation attrb_evaluation_stmt
                      attrb_evaluation_compound_stmt attrb_evaluation_expr
                      attrb_evaluation_stmt_component
                      attrb_evaluation_expr_component
                      attrb_evaluation_stmt_seq
                      attrb_evaluation_stmt_seq_component
%type <name_literal>  named_token nonterminal nonterminal_symbol symbol
                      left_hand_side rhs_symbol
%type <keyword>       array_size_opt nonterminal_declarer
                      precedence_declarer ptr_type_qualifier
                      yacc_directive ox_directive attrb_modifier
%type <userArgSeq>    user_argument_seq user_argument
%type <symDeclSeq>    token_decl_list precedence_decl_list
                      nonterminal_decl_list
%type <typeSeq>       tag_opt tag type_specifier type_list
                      type type_component elaborated_type_name
                      formal_type_specifier abstract_declarator ptr_operator
                      ptr_type_qualifier_seq_opt ptr_type_qualifier_seq
                      direct_abstract_declarator
                      formal_array_qualifier_list_opt
                      formal_array_qualifier_list
                      parameter_list_opt parameter_list
                      parameter_type_list parameter_type
                      type_name formal_array_qualifier
%type <symSeq>        terminal_list token_list token nonterminal_list

%type <keyword>       bison_directive

%type <keyword>       bison_or_byacc_directive

%code
{
#include "config.h"

#include <iomanip>
#include <ostream>
#include <sstream>

#include "bison-directives.hh"
#include "codegen.hh"
#include "ctc.hh"
#include "depGs.hh"
#include "fileMgr.hh"
#include "joinTabs.hh"
#include "mac.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "symTrans.hh"
#include "trav.hh"
#include "treeviz.hh"

// =true if an '@attributes' declaration has occurred
static bool oxAttrbsDecl = false;

// record whether the @autoinh or @autosyn automatic copy rule attribute
// modifier has been seen in the current attribute declaration
static enum attrbDirection autoDef = undefined;

// =true if an '@warn' automatic copy rule attribute modifier has occurred in
// the current attribute declaration
static bool autoWarning = false;

// holds attribute information unique to each 'sort' from the current
// @attributes declaration until ready to add the information to the symbol
// table for each of the grammar symbols of that 'sort'
static sort gSymSort;

// an increasing sequence of sort serial numbers representing the kind of a
// semantic node in the parse tree
static int sortTypeNumSeq = firstSortTypeNum;

// the collection of all defined attribute sorts
static sortCollection sortTable;

// The vector sortSeq holds the list of attributes for the 'sort' (@attributes)
// declaration being processed.  The i-th element of sortSeq is a structure that
// holds the name and copy rule (@autoinh or @autosym, @warn) annotation (or
// default) of the ith attribute of the @attributes declaration being processed.
static sortAttrbSeq sortSeq;

// for checking dynamic traversal modifier restrictions
struct dynTravMod
{
  std::string id;
  int orderRev;            // count the @revorder occurrences
  int dirRev;              // count the @revdirectdion occurrences
  AGC::location dirRevLoc; // @revdirection location (first, if dirRev > 1)

  dynTravMod (std::string i)
      : id (i),
        orderRev (0),
        dirRev (0)
  {}
};

static std::vector<dynTravMod> dynTravModArr;

static int dynTravModSeq;

// =true if "@attributes { ... }" are not all '@void'
static bool hasConcreteType = false;

// checking whether "%destructor { ... } <*>" or "%printer { ... } <*>" occurs
static bool hasDefaultTypedSymbol = false;

#include "agc-Yfile-lexer.hh"

namespace AGC
{
class bisonLexer : public YFILE::Lexer
{
public:

  inline bisonLexer (Yfile_specIO& fR,
#if defined(WITH_REFLEX)
                              const reflex::Input& input
#else // defined(WITH_FLEX)
                              std::ifstream& input
#endif
                             )
      : YFILE::Lexer (fR, input)
  {}

  inline int
  yylex (parser::value_type& yylval, location& yylloc)
  {
    return YFILE::Lexer::yylex (yylval, yylloc);
  }

  inline int
  yylex (parser::value_type* yylval, location* yylloc)
  {
    return YFILE::Lexer::yylex (yylval, yylloc);
  }

  // convert bison -> Y-file parser 'value_type&'
  // lexical tokens in all parser variants are identical
  inline int
  yylex (bison::value_type& yylval, location& yylloc)
  {
    return yylex (reinterpret_cast<parser::value_type&> (yylval), yylloc);
  }

  // convert bison -> Y-file parser 'value_type*'
  // lexical tokens in all parser variants are identical
  inline int
  yylex (bison::value_type* yylval, location* yylloc)
  {
    return yylex (reinterpret_cast<parser::value_type*> (yylval), yylloc);
  }

  inline int
  yylex () // should never be called
  {
    return parser::symbol_kind::S_YYerror;
  }
};
} // namespace AGC

// override Bison default
#define yylex theLexer.yylex
}

%%
AG_spec
  :   parser_spec
  ;

parser_spec
  :   parser_spec_definition_section
      {
        finishYfileDefinitionSection (sortTable);
      }
      "%%"[MARK]
      {
        yoxOutG.append (separatorMARK); // holds the 1st "%%" line
        separatorMARK.clear ();

        // add the 'omitted Lexer return' token & memoize its index
        setupOmittedRetSym ();

        // add the 'error' token & memoize its index
        setupErrorSym ();
      }
      parser_spec_rules_section
      {
        temp1Out.putSolveAndSignalTrailer ();
        temp1Out.putTravCode ();
        temp1Out.putRootAttrbRefSection ();
      }
      tail
      {
        finishYfileRulesSection ();
      }
  ;

parser_spec_definition_section
  :   definition_seq
  ;

parser_spec_rules_section
  :   rule_seq
  ;

tail
  :   %empty
      {
        yoxOutG.gen ().put ("\n%%\n");
      }
  |   "%%"
      {
        yoxOutG.append (separatorMARK); // holds the 2nd "%%" line
        separatorMARK.clear ();
      }
  ;

definition_seq[left]
  :   %empty
      {
        @left = yyla.location;
      }
  |   definition_seq[right] definition
  ;

definition
  :   yacc_directive
      {
        if (oxAttrbsDecl)
          {
            agc_syntax_error (@yacc_directive,
                              "syntax error, "
                              "\"" + $yacc_directive + "\""
                              " directive follows an Ox declaration.");
            yyerrcnt++;
          }
      }
  |   "%{" code_seq "%}"
  |   bison_or_byacc_directive
  |   bison_directive
  |   ox_directive
      {
        oxAttrbsDecl = true;
      }
  |   error
      {
        yyclearin;
      }
  ;

yacc_directive
  :   "%start"[START] "nonterminal symbol"[GS_ID]
      {
        yoxOutGdefs.gen ().put (yoxPrefix (augNstr));

        installStartSym (@GS_ID, $GS_ID);

        $yacc_directive = $START;
      }
  |   "%union"[UNION] unionTag_opt
      {
        if (yaccUnionDefined++ == 0)
          yoxOutGdefs.gen ()
              .put ("/* `%union\' declaration was here */\n");
        else
          yoxOutGdefs.gen ()
              .put ("/* `%union\' "
                    "#" + std::to_string (yaccUnionDefined) +
                    " declaration was here */\n");
      }
      "{"
      // the body of the '%union' declaration is (currently) passed thru by the lexer
      "}"
      {
        if (yaccUnionDefined > 1)
          multipleYYSTYPEdefinitions (@yacc_directive, "%union");

        if (bisonApiValueTypeUnionDefined)
          conflictingYYSTYPEdefinitions (@yacc_directive,
                                         "%union",
                                         "%define api.value.type");

        $yacc_directive = $UNION;
      }
  |   "%token"[TOKEN] token_decl_list
      {
        const symDeclSeq& token_decl_list = $token_decl_list;

        try
          {
            global_symDeclSeq.insert (global_symDeclSeq.end (),
                                      token_decl_list.cbegin (),
                                      token_decl_list.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $yacc_directive = $TOKEN;
      }
  |   precedence_declarer precedence_decl_list
      {
        const symDeclSeq& precedence_decl_list = $precedence_decl_list;

        try
          {
            global_symDeclSeq.insert (global_symDeclSeq.end (),
                                      precedence_decl_list.cbegin (),
                                      precedence_decl_list.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $yacc_directive = $precedence_declarer;
      }
  |   nonterminal_declarer nonterminal_decl_list
      {
        const symDeclSeq& nonterminal_decl_list = $nonterminal_decl_list;

        try
          {
            global_symDeclSeq.insert (global_symDeclSeq.end (),
                                      nonterminal_decl_list.cbegin (),
                                      nonterminal_decl_list.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $yacc_directive = $nonterminal_declarer;
      }
  ;

unionTag_opt
  :   %empty
  |   "identifier"[ID]
      {
        ctrl.unionTag = $ID;
      }
  ;

bison_or_byacc_directive
  :   "%destructor"[DESTRUCTOR] code_block
      {
        hasDefaultTypedSymbol = false;
      }
      destructor_symbol_list
      {
        if (hasDefaultTypedSymbol)
          needDestructorForStartSymbol = true;

        $bison_or_byacc_directive = $DESTRUCTOR;
      }
  |   "%expect"[EXPECT] "number"
      {
        $bison_or_byacc_directive = $EXPECT;
      }
  |   "%expect-rr"[EXPECT_RR] "number"
      {
        $bison_or_byacc_directive = $EXPECT_RR;
      }
  |   "%lex-param"[LEX_PARAM] user_argument_seq
      {
        $bison_or_byacc_directive = $LEX_PARAM;
      }
  |   "%parse-param"[PARSE_PARAM] user_argument_seq
      {
        const userArgSeq& user_argument_seq = $user_argument_seq;

        try
          {
            global_userArgSeq.insert (global_userArgSeq.end (),
                                      user_argument_seq.cbegin (),
                                      user_argument_seq.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $bison_or_byacc_directive = $PARSE_PARAM;
      }
  |   "%pure-parser"[PURE_PARSER]
      {
        ctrl.allLfilesForReentrantParser = true;

        $bison_or_byacc_directive = $PURE_PARSER;
      }
  |   CODE code_block
      {
        $bison_or_byacc_directive = $CODE;
      }
  ;

bison_directive
  :   "%define"[DEFINE] api_param_def
      {
        $bison_directive = $DEFINE;
      }
  |   "%name-prefix"[API_VARIABLE] LITERAL_STRING
      {
        API_variable_LITERAL (@bison_directive, $API_VARIABLE, $LITERAL_STRING);

        $bison_directive = $API_VARIABLE;
      }
  |   "%param"[PARAM] user_argument_seq
      {
        const userArgSeq& user_argument_seq = $user_argument_seq;

        try
          {
            global_userArgSeq.insert (global_userArgSeq.end (),
                                      user_argument_seq.cbegin (),
                                      user_argument_seq.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $bison_directive = $PARAM;
      }
  |   "%language"[LANGUAGE] LITERAL_STRING
      {
        language_directive (@bison_directive, $LITERAL_STRING);

        $bison_directive = $LANGUAGE;
      }
  |   "%printer"[PRINTER] code_block
      {
        hasDefaultTypedSymbol = false;
      }
      destructor_symbol_list
      {
        if (hasDefaultTypedSymbol)
          needPrinterForStartSymbol = true;

        $bison_directive = $PRINTER;
      }
  |   "%skeleton"[SKELETON] LITERAL_STRING
      {
        skeleton_directive (@bison_directive, $LITERAL_STRING);

        $bison_directive = $SKELETON;
      }
  ;

api_param_def // Bison "%define" directive parameters
  :   "bison API variable"[API_VARIABLE]
      {
        API_variable_default (@api_param_def, $API_VARIABLE);
      }
  |   "bison API variable"[API_VARIABLE] "bison API value"[ID]
      {
        API_variable_ID (@api_param_def, $API_VARIABLE, $ID);
      }
  |   "bison API variable"[API_VARIABLE] LITERAL_STRING
      {
        API_variable_LITERAL (@api_param_def, $API_VARIABLE, $LITERAL_STRING);
      }
  |   "bison API variable"[API_VARIABLE] "{" LITERAL_STRING "}"
      {
        API_variable_LITERAL (@api_param_def, $API_VARIABLE, $LITERAL_STRING);
      }
  |   "bison API variable"[API_VARIABLE] "{" LITERAL_PATH "}"
      {
        API_variable_LITERAL (@api_param_def, $API_VARIABLE, $LITERAL_PATH);
      }
  |   "bison API variable"[API_VARIABLE] "{" "}"
  |   "bison API variable"[API_VARIABLE] "{" formal_type_specifier "}"
      {
        const name_literal& apivar = $API_VARIABLE;

        int num_values = API_variable_EXPR (@api_param_def,
                                            apivar,
                                            $formal_type_specifier);
        if (num_values > 1)
          agc_syntax_error (@api_param_def,
                            "syntax error, " + apivar + " is single-valued "
                            "but " + std::to_string (num_values) + " values "
                            "were provided");
      }
  ;

ox_directive
  :   {
        // Y-file lexer hack: suppress copying attribute names onto the output
        // file when processing a global automatic copy rule attribute modifier
        // declaration
        ctrl.process_autoSynOrInh = true;
      }
      attrb_modifier autoIDseq
      {
        // reset globals for the next autoIDseq
        autoDef = undefined;
        autoWarning = false;
        ctrl.process_autoSynOrInh = false;

        $ox_directive = $attrb_modifier;
        @ox_directive.begin = @attrb_modifier.begin;
      }
  |   "@attributes"[OX_ATTRIBUTES] "{" attribute_decl_list "}"
      {
        gSymSort.typeNum = sortTypeNumSeq; // sort serial #

        if (hasConcreteType)
          {
            // create a unique C/C++ type name for the new sort
            gSymSort.typeName =
                yoxPrefix ("T") + std::to_string (sortTypeNumSeq);
          }

        sortTypeNumSeq++;

        yoxOutHrequires.putAttrbTypeDecl (gSymSort.typeNum,
                                          gSymSort.typeName,
                                          hasConcreteType);

        try
          {
            // Allocate a permanent list of attributes for the completed
            // @attributes declaration, to be pointed to by the grammar symbol
            // table entry of each grammar symbol of the sort.
            gSymSort.attrbSeq = new sortAttrbSeq (sortSeq);
          }
        catch (std::bad_alloc const&)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        gSymSort.nAttrbs = sortSeq.size ();

        try
          {
            // keep track of all of the attribute sorts
            sortTable.push_back (gSymSort);
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        putAttrbSort (gSymSort);
      }
      attributed_symbol_list
      {
        // reset globals for the next @attributes declaration
        gSymSort = sort ();      // undefined
        sortSeq.clear ();        // empty
        hasConcreteType = false; // all attributes are '@void'

        $ox_directive = $OX_ATTRIBUTES;
      }
  |   "@macro"[OX_MACRO] "identifier"[ID]
      {
        if (!initMacroDef ($ID))
          yox_error (@ID, "macro multiply defined: \"" + $ID + "\".\n");
      }
      "(" macroFormalList ")"
      // the body of the macro is (currently) handled by the lexer
      "@end"
      {
        $ox_directive = $OX_MACRO;
      }
  |   "@traversal"[OX_TRAV]
      {
        clearTravBuf ();
      }
      travSpecSeq
      {
        if (travSpecConflict ())
          yox_error (@travSpecSeq,
                     "contradictory traversal declaration.\n");
      }
      travIDseq
      {
        // suppress pruning if any @traversal is enabled
        ctrl.suppressPruning |= (!travDisabled ());

        $ox_directive = $OX_TRAV;
      }
  ;

code_seq[left] // verbatim code is (mostly) passed thru by the lexer, save for
               // '{' ... '}' (C/C++ compound statement braces) in 'code_block'
  :   %empty
      {
        @left = yyla.location;
      }
  |   code_seq[right] code_block
  ;

token_decl_list[left]
  :   tag_opt token_list
      {
        const typeSeq& tag_opt = $tag_opt;

        if (tag_opt.empty ())
          $left = symDeclSeq ();
        else
          $left = symDeclSeq (1, symDeclSeqElem (tag_opt, $token_list));
      }
  |   token_decl_list[right] tag token_list
      {
        symDeclSeq new_seq = $right;

        try
          {
            new_seq.push_back (symDeclSeqElem ($tag, $token_list));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

tag_opt
  :   %empty
      {
        @tag_opt = yyla.location;
        $tag_opt = typeSeq (); // empty sequence
      }
  |   tag
      {
        $tag_opt = $tag;
      }
  ;

tag
  :   "<" formal_type_specifier ">"
      {
        // Bison expects a C/C++ type if "%define api.value.type variant",
        // however Ox doesn't currently support code generation for this
        $tag = $formal_type_specifier;
      }
  ;

token_list[left]
  :   token
      {
        $left = $token;
      }
  |   token_list[right] token
      {
        symSeq new_seq = $right;

        const symSeq& token = $token;

        try
          {
            new_seq.insert (new_seq.end (), token.cbegin (), token.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  |   token_list[right] "," token
      {
        symSeq new_seq = $right;

        const symSeq& token = $token;

        try
          {
            new_seq.insert (new_seq.end (), token.cbegin (), token.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

token
  :   named_token
      {
        // Yacc will report an error if 'named_token' is a nonterminal
        const std::string& named_token = $named_token;

        installGsymAsTokenIfNew (@named_token, named_token);
        $token = symSeq (1, named_token);
      }
  |   LITERAL_CHAR
      {
        const std::string& LITERAL_CHAR = $LITERAL_CHAR;

        installGsymAsTokenIfNew (@LITERAL_CHAR, LITERAL_CHAR);
        $token = symSeq (1, LITERAL_CHAR);
      }
  |   named_token LITERAL_STRING
      {
        // Bison will report an error if 'named_token' is a nonterminal
        const std::string& named_token = $named_token;

        installGsymAsTokenIfNew (@named_token, named_token);

        const std::string& LITERAL_STRING = $LITERAL_STRING;

        installAliasToToken (@LITERAL_STRING, LITERAL_STRING,
                             @named_token, named_token);

        symSeq new_seq = symSeq (1, named_token);

        try
          {
            new_seq.push_back (LITERAL_STRING);
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $token = new_seq;
      }
  ;

named_token
  :   "identifier"[ID] NUM_opt
      {
        $named_token = $ID;
      }
  ;

NUM_opt
  :   %empty
  |   "number"
  ;

precedence_declarer
  :   "%left"[LEFT]
      {
        $precedence_declarer = $LEFT;
      }
  |   "%right"[RIGHT]
      {
        $precedence_declarer = $RIGHT;
      }
  |   "%nonassoc"[NONASSOC]
      {
        $precedence_declarer = $NONASSOC;
      }
  |   "%precedence"[PRECEDENCE]
      {
        $precedence_declarer = $PRECEDENCE;
      }
  ;

precedence_decl_list[left]
  :   tag_opt terminal_list
      {
        const typeSeq& tag_opt = $tag_opt;

        if (tag_opt.empty ())
          $left = symDeclSeq ();
        else
          $left = symDeclSeq (1, symDeclSeqElem (tag_opt, $terminal_list));
      }
  |   precedence_decl_list[right] tag terminal_list
      {
        symDeclSeq new_seq = $right;

        try
          {
            new_seq.push_back (symDeclSeqElem ($tag, $terminal_list));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

terminal_list[left]
  :   terminal_symbol
      {
        $left = symSeq (1, $terminal_symbol);
      }
  |   terminal_list[right] terminal_symbol
      {
        symSeq new_seq = $right;

        try
          {
            new_seq.push_back ($terminal_symbol);
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
    }

        $left = new_seq;
      }
  |   terminal_list[right] "," terminal_symbol
      {
        symSeq new_seq = $right;

        try
          {
            new_seq.push_back ($terminal_symbol);
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

terminal_symbol
  :   named_token
      {
        // Yacc will report an error if 'named_token' is a nonterminal
        installGsymAsTokenIfNew (@named_token, $named_token);
      }
  |   LITERAL_CHAR
      {
        installGsymAsTokenIfNew (@LITERAL_CHAR, $LITERAL_CHAR);
      }
  |   LITERAL_STRING
      {
        installGsymAsTokenIfNew (@LITERAL_STRING, $LITERAL_STRING);
      }
  ;

nonterminal_symbol
  :   "nonterminal symbol"[GS_ID]
      {
        $nonterminal_symbol = $GS_ID;
      }
  |   "identifier"[ID]
      {
        $nonterminal_symbol = $ID;
      }
  ;

symbol
  :   "nonterminal symbol"[GS_ID]
      {
        $symbol = $GS_ID;
      }
  |   "identifier"[ID]
      {
        $symbol = $ID;
      }
  |   LITERAL_CHAR
      {
        $symbol = $LITERAL_CHAR;
      }
  |   LITERAL_STRING
      {
        $symbol = $LITERAL_STRING;
      }
  ;

nonterminal_declarer
  :   "%type"[TYPE]
      {
        $nonterminal_declarer = $TYPE;
      }
  |   "%nterm"[NTERM]
      {
        $nonterminal_declarer = $NTERM;
      }
  ;

nonterminal_decl_list[left]
  :   tag_opt nonterminal_list
      {
        const typeSeq& tag_opt = $tag_opt;

        if (tag_opt.empty ())
          $left = symDeclSeq ();
        else
          $left = symDeclSeq (1, symDeclSeqElem (tag_opt, $nonterminal_list));
      }
  |   nonterminal_decl_list[right] tag nonterminal_list
      {
        symDeclSeq new_seq = $right;

        try
          {
            new_seq.push_back (symDeclSeqElem ($tag, $nonterminal_list));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

nonterminal_list[left]
  :   nonterminal
      {
        $left = symSeq (1, $nonterminal);
      }
  |   nonterminal_list[right] nonterminal
      {
        symSeq new_seq = $right;

        try
          {
            new_seq.push_back ($nonterminal);
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  |   nonterminal_list[right] "," nonterminal
      {
        symSeq new_seq = $right;

        try
          {
            new_seq.push_back ($nonterminal);
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

nonterminal
  :   nonterminal_symbol
      {
        $nonterminal = $nonterminal_symbol;
      }
  ;

code_block
  :   "{" code_seq "}"
  |   "{" error "}"
  ;

user_argument_seq[left]
  :   user_argument
  |   user_argument_seq[right] user_argument
      {
        userArgSeq new_seq = $right;

        const userArgSeq& user_argument = $user_argument;

        try
          {
            new_seq.insert (new_seq.end (),
                            user_argument.cbegin (),
                            user_argument.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

user_argument
  :   "{" formal_type_specifier "identifier"[ID] "}"
      {
        $user_argument =
            userArgSeq (1, userArgSeqElem ($formal_type_specifier, $ID));
      }
  ;

destructor_symbol_list
  :   destructor_symbol
  |   destructor_symbol_list destructor_symbol
  ;

destructor_symbol[left]
  :   symbol
  |   tag
  |   "<" "*" ">"
      {
        hasDefaultTypedSymbol = true;
      }
  |   "<" ">"
  ;

autoIDseq
  :   autoID
  |   autoIDseq autoID
  ;

autoID
  :   "identifier"[ID]
      {
        attrbAutoDef (@ID, $ID, autoDef, autoWarning);
      }
  ;

attribute_decl_list[left]
  :   attribute_decl
  |   attribute_decl_list[right] attribute_decl
  ;

attribute_decl
  :   attrb_modifier_opt type_specifier id_seq ";"
      {
        hasConcreteType = true; // at least one attribute is not '@void'

        // reset globals for the next autoIDseq
        autoDef = undefined;
        autoWarning = false;
      }
  |   "@void" id_seq2 ";"
  |   error ";"
  ;

attrb_modifier_opt
  :   %empty
      {
        @attrb_modifier_opt = yyla.location;

        // no automatic copy rule attribute modifier declaration
        autoDef = undefined;
        autoWarning = false;
      }
  |   attrb_modifier
  ;

attrb_modifier
  :   "@autosyn"[OX_AUTOSYN]
      {
        switch (autoDef)
          {
          case inherited:
            yox_error (@OX_AUTOSYN,
                       "@autosyn conflicts with preceding @autoinh.\n");
            break;

          case synthesized:
            yox_warning (@OX_AUTOSYN, "repeated @autosyn (ignored).\n");
            break;

          case undefined:
            autoDef = synthesized;
            break;
          }
      }
      atWarn_opt
      {
        $attrb_modifier = $OX_AUTOSYN;
      }
  |   "@autoinh"[OX_AUTOINH]
      {
        switch (autoDef)
          {
          case inherited:
            yox_warning (@OX_AUTOINH, "repeated @autoinh (ignored).\n");
            break;

          case synthesized:
            yox_error (@OX_AUTOINH,
                       "@autoihn conflicts with preceding @autosyn.\n");
            break;

          case undefined:
            autoDef = inherited;
            break;
          }
      }
      atWarn_opt
      {
        $attrb_modifier = $OX_AUTOINH;
      }
  ;

atWarn_opt
  :   %empty
      {
        autoWarning = false;
      }
  |   "@warn"
      {
        autoWarning = true;
      }
  ;

type_specifier
  :   type %expect 1
      {
        $type_specifier = $type;
      }
  |   type // syntax error for type_keyword(s) here
      "<" type_list ">" // C++
      {
        typeSeq new_seq = $type;

        const typeSeq& type_list = $type_list;

        try
          {
            new_seq.push_back (typeSeqElem ("<"));

            new_seq.insert (new_seq.end (),
                            type_list.cbegin (),
                            type_list.cend ());

            new_seq.push_back (typeSeqElem (">"));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $type_specifier = new_seq;
      }
  ;

type[left]
  :   type_component
      {
       $left = $type_component;
      }
  |   type[right] type_component
      {
        typeSeq new_seq = $right;

        const typeSeq& type_component = $type_component;

        try
          {
            new_seq.insert (new_seq.end (),
                            type_component.cbegin (),
                            type_component.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

type_component
  :   TYPE_KEYWORD
      {
        $type_component =
            typeSeq (1, typeSeqElem ($TYPE_KEYWORD, is_TYPE_KEYWORD));
      }
  |   "const"[CONST]
      {
        $type_component = typeSeq (1, typeSeqElem ($CONST, is_TYPE_KEYWORD));
      }
  |   "volatile"[VOLATILE]
      {
        $type_component = typeSeq (1, typeSeqElem ($VOLATILE, is_TYPE_KEYWORD));
      }
  |   type_name
      {
        $type_component = $type_name;
      }
  |   elaborated_type_name
      {
        $type_component = $elaborated_type_name;
      }
  ;

type_name
  :   "identifier"[ID]
      {
        $type_name = typeSeq (1, typeSeqElem ($ID, is_ID));
      }
  |   NAMESPACE_ID // C++
      {
        $type_name = typeSeq (1, typeSeqElem ($NAMESPACE_ID, is_NAMESPACE_ID));
      }
  ;

elaborated_type_name
  :   ELABORATED_TYPE_KEYWORD type_name
      {
        typeSeq new_seq =
            typeSeq (1, typeSeqElem ($ELABORATED_TYPE_KEYWORD,
                                     is_TYPE_KEYWORD));

        const typeSeq& type_name = $type_name;

        try
          {
            new_seq.insert (new_seq.end (),
                            type_name.cbegin (),
                            type_name.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $elaborated_type_name = new_seq;
      }
  ;

type_list[left] // C++
  :   type
      {
        $left = $type;
      }
  |   type_list[right] "," type
      {
        typeSeq new_seq = $right;

        const typeSeq& type = $type;

        try
          {
            new_seq.push_back (typeSeqElem (","));

            new_seq.insert (new_seq.end (),
                            type.cbegin (),
                            type.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

id_seq[left]
  :   declarator
  |   id_seq[right] "," declarator
  ;

declarator[left]
  :   direct_declarator field_length_spec
  |   ptr_operator declarator[right]
  ;

direct_declarator[left]
  :   attribute_id
  |   "(" declarator ")"
  |   direct_declarator[right] "[" array_size_opt "]"
  |   direct_declarator[right] "(" parameter_list_opt ")"
  ;

id_seq2[left]
  :   attribute_id
  |   id_seq2[right] "," attribute_id
  ;

attribute_id
  :   "identifier"[ID]
      {
        const sortAttrb attribute_id =
            newAttrbID (@ID, $ID, autoDef, autoWarning);

        bool add_OK = true;

        // The original order of appearance of each attribute in the
        // declaration list is preserved, with the exception that
        // duplicates are deleted.

        // check for duplicated attribute name in the sort
        const std::string& attrbName = attribute_id.id;

        for (const sortAttrb& attrb : sortSeq)
          if (attrb.id == attrbName)
            {
              yox_error (@ID,
                         "duplicated attribute: \"" + attrbName + "\""
                         ".\n");
              add_OK = false;
              break;
            }

        if (add_OK)
          try
            {
              // if attribute isn't duplicate, add it to the list
              sortSeq.push_back (attribute_id);
            }
          catch (...)
            {
              yox_alloc_fail (__FILE__, __LINE__);
            }
      }
  ;

ptr_operator
  :   "*"[ASTERISK] ptr_type_qualifier_seq_opt
      {
        typeSeq new_seq = typeSeq (1, typeSeqElem ("*"));

        const typeSeq& ptr_type_qualifier_seq_opt = $ptr_type_qualifier_seq_opt;

        try
          {
            new_seq.insert (new_seq.end (),
                            ptr_type_qualifier_seq_opt.cbegin (),
                            ptr_type_qualifier_seq_opt.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $ptr_operator = new_seq;
      }
  |   "&"
      {
        $ptr_operator = typeSeq (1, typeSeqElem ("&"));
      }
  ;

ptr_type_qualifier_seq_opt
  :   %empty
      {
        $ptr_type_qualifier_seq_opt = typeSeq (); // empty sequence
      }
  |   ptr_type_qualifier_seq
      {
        $ptr_type_qualifier_seq_opt = $ptr_type_qualifier_seq;
      }
  ;

ptr_type_qualifier_seq[left]
  :   ptr_type_qualifier
      {
        $left = typeSeq (1, typeSeqElem ($ptr_type_qualifier, is_TYPE_KEYWORD));
      }
  |   ptr_type_qualifier_seq[right] ptr_type_qualifier
      {
        typeSeq new_seq = $right;

        try
          {
            new_seq.push_back (typeSeqElem ($ptr_type_qualifier,
                                            is_TYPE_KEYWORD));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

ptr_type_qualifier
  :   "const"[CONST]
      {
        $ptr_type_qualifier = $CONST;
      }
  |   "restrict"[RESTRICT] // C99
      {
        $ptr_type_qualifier = $RESTRICT;
      }
  ;

field_length_spec
  :   %empty
  |   ":" "number"
  ;

array_size_opt
  :   %empty
      {
        @array_size_opt = yyla.location;
        $array_size_opt = nullptr;
      }
  |   "number"[NUM]
      {
        $array_size_opt = $NUM.spelling;
      }
  |   "*" // C99
      {
        // treat it like a 'type_keyword' for the moment
        $array_size_opt = const_cast<char*> ("*");
      }
  |   "identifier"[ID]
      {
        // treat it like a 'type_keyword' for the moment
        $array_size_opt = $ID;
      }
  ;

parameter_list_opt
  :   %empty
      {
        $parameter_list_opt = typeSeq (); // empty sequence
      }
  |   parameter_list
      {
        $parameter_list_opt = $parameter_list;
      }
  ;

parameter_list
  :   parameter_type_list
      {
        $parameter_list = $parameter_type_list;
      }
  |   parameter_type_list "," "..."
      {
        typeSeq new_seq = $parameter_type_list;

        try
          {
            new_seq.push_back (typeSeqElem (","));
            new_seq.push_back (typeSeqElem ("..."));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $parameter_list = new_seq;
      }
  ;

parameter_type_list[left]
  :   parameter_type
      {
        $left = $parameter_type;
      }
  |   parameter_type_list[right] "," parameter_type
      {
        typeSeq new_seq = $right;

        const typeSeq& parameter_type = $parameter_type;

        try
          {
            new_seq.push_back (typeSeqElem (","));

            new_seq.insert (new_seq.end (),
                            parameter_type.cbegin (),
                            parameter_type.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

parameter_type
  :   formal_type_specifier
      {
        $parameter_type = $formal_type_specifier;
      }
  |   "register"[REGISTER] formal_type_specifier
      {
        typeSeq new_seq = typeSeq (1, typeSeqElem ($REGISTER, is_TYPE_KEYWORD));

        const typeSeq& formal_type_specifier = $formal_type_specifier;

        try
          {
            new_seq.insert (new_seq.end (),
                            formal_type_specifier.cbegin (),
                            formal_type_specifier.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $parameter_type = new_seq;
      }
  ;

formal_type_specifier
  :   type_specifier abstract_declarator
      {
        typeSeq new_seq = $type_specifier;

        const typeSeq& abstract_declarator = $abstract_declarator;

        try
          {
            new_seq.insert (new_seq.end (),
                            abstract_declarator.cbegin (),
                            abstract_declarator.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $formal_type_specifier = new_seq;
      }
  ;

abstract_declarator[left]
  :   %empty
      {
        $left = typeSeq (); // empty sequence
      }
  |   direct_abstract_declarator
      {
        $left = $direct_abstract_declarator;
      }
  |   ptr_operator abstract_declarator[right]
      {
        typeSeq new_seq = $ptr_operator;

        const typeSeq& right = $right;

        try
          {
            new_seq.insert (new_seq.end (), right.cbegin (), right.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

direct_abstract_declarator[left]
  :   "(" abstract_declarator ")"
      {
        typeSeq new_seq = typeSeq (1, typeSeqElem ("(" /*)*/));

        const typeSeq& abstract_declarator = $abstract_declarator;

        try
          {
            new_seq.insert (new_seq.end (),
                            abstract_declarator.cbegin (),
                            abstract_declarator.cend ());

            new_seq.push_back (typeSeqElem (/*(*/ ")"));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  |   direct_abstract_declarator[right]
      "[" formal_array_qualifier_list_opt array_size_opt "]"
      {
        typeSeq new_seq = $right;

        try
          {
            new_seq.push_back (typeSeqElem ("["));

            const typeSeq& formal_array_qualifier_list_opt =
                $formal_array_qualifier_list_opt;

            new_seq.insert (new_seq.end (),
                            formal_array_qualifier_list_opt.cbegin (),
                            formal_array_qualifier_list_opt.cend ());

            const std::string& array_size = $array_size_opt;

            if (!array_size.empty ())
              new_seq.push_back (typeSeqElem (array_size));

            new_seq.push_back (typeSeqElem ("]"));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  |   direct_abstract_declarator[right] "(" parameter_list_opt ")"
      {
        typeSeq new_seq = $right;

        try
          {
            new_seq.push_back (typeSeqElem ("(" /*)*/));

            const typeSeq& parameter_list_opt = $parameter_list_opt;

            new_seq.insert (new_seq.end (),
                            parameter_list_opt.cbegin (),
                            parameter_list_opt.cend ());

            new_seq.push_back (typeSeqElem (/*(*/ ")"));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;


formal_array_qualifier_list_opt // C99
  :   %empty
      {
        @formal_array_qualifier_list_opt = yyla.location;
        $formal_array_qualifier_list_opt = typeSeq (); // empty sequence
      }
  |   formal_array_qualifier_list
      {
        $formal_array_qualifier_list_opt = $formal_array_qualifier_list;
      }
  ;

formal_array_qualifier_list[left] // C99
  :   formal_array_qualifier
      {
        $left = $formal_array_qualifier;
      }
  |   formal_array_qualifier_list[right] formal_array_qualifier
      {
        typeSeq new_seq = $right;

        const typeSeq& formal_array_qualifier = $formal_array_qualifier;

        try
          {
            new_seq.insert (new_seq.end (),
                            formal_array_qualifier.cbegin (),
                            formal_array_qualifier.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  ;

formal_array_qualifier // C99
  :   "const"[CONST]
      {
        $formal_array_qualifier =
            typeSeq (1, typeSeqElem ($CONST, is_TYPE_KEYWORD));
      }
  |   "volatile"[VOLATILE]
      {
        $formal_array_qualifier =
            typeSeq (1, typeSeqElem ($VOLATILE, is_TYPE_KEYWORD));
      }
  |   "restrict"[RESTRICT]
      {
        $formal_array_qualifier =
            typeSeq (1, typeSeqElem ($RESTRICT, is_TYPE_KEYWORD));
      }
  |   "static"[STATIC]
      {
        $formal_array_qualifier =
            typeSeq (1, typeSeqElem ($STATIC, is_TYPE_KEYWORD));
      }
  ;

attributed_symbol_list
  :   attributed_symbol
  |   attributed_symbol_list attributed_symbol
  ;

attributed_symbol
  :   symbol
      {
        std::string symbol = $symbol;

        installGsymIfNew (@symbol, symbol);
        // if 'symbol' is an alias, 'installGsymIfNew ()' replaced it
        // with the base symbol

        addAttrbsToGsym (@symbol, symbol, gSymSort);
      }
  ;

macroFormalList
  :   "identifier"[ID]
      {
        (void)macro.installMacroFormal ($ID);
      }
  |   macroFormalList "," "identifier"[ID]
      {
        const std::string& ID = $ID;
        if (!macro.installMacroFormal (ID))
          yox_error (@ID,
                     "duplicated macro formal parameter: "
                     "\"" + ID + "\".\n");
      }
  ;

travSpecSeq[left]
  :   %empty
  |   travSpecSeq[right] travSpec
  ;

travIDseq
  :   travID
  |   travIDseq travID
  ;

travID
  :   "identifier"[ID]
      {
        newTrav (@ID, $ID);
      }
  ;

travSpec
  :   "@preorder"
      {
        travSetPreorder ();
      }
  |   "@postorder"
      {
        travSetPostorder ();
      }
  |   "@lefttoright"
      {
        travSetLeftToRight ();
      }
  |   "@righttoleft"
      {
        travSetRightToLeft ();
      }
  |   "@disable"
      {
        travSetDisable ();
      }
  ;

rule_seq
  :   rule_set ";"
  |   rule_seq rule_set ";"
  ;

rule_set[RULE]
  :   left_hand_side ":"[COLON]
      {
        const std::string& left_hand_side = $left_hand_side;

        static bool alreadyAugmented = false;

        if (!alreadyAugmented)
          {
            if (startSymLexeme.empty ())
              installStartSym (@left_hand_side, left_hand_side);

            // Yacc pseudo variable '$1' is the original user-defined
            // start symbol; use '$<yyyOxAttrbs>1' since the Yacc
            // semantic type may be unknown

            const std::string rootNodeReference =
                "$<" + yoxPrefix ("OxAttrbs") + ">1.semEl->node";

            yoxOutG.gen ()
                .put ("\n" + yoxPrefix (augNstr) + "\n"
                      "\t:\t\t" + startSymLexeme + "\n"
                      "\t\t{if(" + yoxPrefix ("Yok") + ")\n"
                      "\t\t   {").putDecorateCall ("true") // -> pruning
                .put ("\t\t    " +
                      yoxPrefix ("ExecuteRRsection") +
                      "(" + rootNodeReference + ");\n")
                .putDoTraversals (rootNodeReference)
                .put ("\t\t   }\n"
                      "\t\t}\n\t;\n");

            alreadyAugmented = true;
          }

        yoxOutG.append (possibleLHS); // dump the pending LHS text
        possibleLHS.clear ();

        prodBufferPointer2 = 0;
        incProdNumSeq ();

        yaccSemIndex = 0; // 0 for LHS symbol; ignored

        installGsymAsNontermIfNew (@left_hand_side, left_hand_side);

        putSymbolInProdBuffer2 (left_hand_side);
      }
      right_hand_side
  |   rule_set "|"[OR]
      {
        prodBufferPointer2 = 1; // keep LHS from prior rule
        incProdNumSeq ();

        yaccSemIndex = 1; // 1st RHS symbol
      }
      right_hand_side
  ;

left_hand_side
  :   nonterminal_symbol
      {
        $left_hand_side = $nonterminal_symbol;
      }
  |   nonterminal_symbol named_reference_def
      {
        $left_hand_side = $nonterminal_symbol;
      }
  ;

right_hand_side
  :   rhs_body
      {
        initLDG ();
      }
      attrb_ref_section
      {
        finishYfileAttrbRefSection (@right_hand_side);
      }
  ;

rhs_body
  :   %empty
      {
        @rhs_body = yyla.location;
      }
  |   rhs_seq
  |   "%empty"
  |   "%empty" rhs_seq // overly general; meh!
  ;

rhs_seq
  :   rhs_component
  |   rhs_seq rhs_component
  ;

rhs_component
  :   rhs_item
  |   rule_directive
  |   parser_action
      {
        // Yacc will generate a nonterminal here and a corresponding empty rule
        // with the semantic action for that nonterminal; Ox semantic action
        // processing doesn't know about this rule, and so doesn't know about
        // this entry on the Yacc semantic stack; this offset calculation keeps
        // track of where RHS symbols occur in the Yacc semantic stack

        yaccSemIndex++; // account for the fake RHS symbol
      }
  |   error
      {
        yyclearin;
      }
  ;

rhs_item
  :   rhs_symbol
      {
        std::string rhs_symbol = $rhs_symbol;

        installGsymIfNew (@rhs_symbol, rhs_symbol);
        // if 'rhs_symbol' is an alias, 'installGsymIfNew ()' replaced it
        // with the base symbol

        putSymbolInProdBuffer2 (rhs_symbol);
      }
  ;

rhs_symbol
  :   symbol
      {
        $rhs_symbol = $symbol;
      }
  |   symbol named_reference_def
      {
        $rhs_symbol = $symbol;
      }
  ;

named_reference_def
  :   "[" "identifier" "]"
  ;

rule_directive
  :   "%prec"[PREC] symbol
  |   "%dprec"[DPREC] "number"
  |   "%expect"[EXPECT] "number"
  |   "%expect-rr"[EXPECT_RR] "number"
  |   "%merge"[MERGE] "<" "identifier" ">"
      {
        notSupported (@MERGE, "\"%merge\"");
      }
  ;

parser_action
  :   code_block
  |   bison_action
  ;

bison_action
  :   tag // the midrule action semantic type
      code_block
  |   code_block named_reference_def
  |   tag // the midrule action semantic type
      code_block named_reference_def
  |   "%?" code_block // predicate action
  ;

attrb_ref_section
  :   %empty
      {
        @attrb_ref_section = AGC::location (@-1.end); // @rhs_body.end
      }
  |   "@{"
      {
        // setup for checking dynamic traversal modifier restrictions
        dynTravModSeq = -1;
        dynTravModArr.clear ();

        decorationRequested = false; // default: no decoration
      }
      attrb_ref_section_body
      "@}"
      {
        // if @revdirection occurs in multiple actions, the actions must have
        // different annunciators

        for (int i = 0; i <= dynTravModSeq; i++)
          {
            for (int j = i + 1; j <= dynTravModSeq; j++)
              {
                if (dynTravModArr[i].id == dynTravModArr[j].id &&
                    dynTravModArr[i].dirRev > 0 &&
                    dynTravModArr[j].dirRev > 0)
                  {
                    if (dynTravModArr[dynTravModSeq].id != errTravID)
                      yox_error (dynTravModArr[j].dirRevLoc,
                                 "multiple @revdirection modifiers for "
                                 "traversal "
                                 "\"" + dynTravModArr[j].id + "\".\n");
                  }
              }
          }
      }
  ;

attrb_ref_section_body
  :   %empty
  |   attrb_ref_section_seq
  |   decorate_cmd
  |   decorate_cmd attrb_ref_section_seq
  ;

decorate_cmd
  :   "@decorate"
      {
        decorationRequested = true;
      }
  ;

attrb_ref_section_seq
  :   attrb_ref_section_item
  |   attrb_ref_section_seq attrb_ref_section_item
  ;

attrb_ref_section_item
  :   attrb_def
      {
        processAdepSpec ();
      }
  |   traversal_segment
  |   /* empty */ ";"
  |   error
      {
        yyclearin;
      }
  ;

// attribute definition mode

attrb_def
  :   "@i" attrb_evaluation
      {
        const attrbRefSeq& attrbRefs = $attrb_evaluation;

        if (attrbRefs.empty ()) // no attribute occurrences
          yox_warning (@attrb_evaluation,
                       "attribute definition doesn\'t define an attribute");
        else
          {
            // find the first non-NIO attribute occurrence => the dependent
            attrbRefSeq::const_iterator it = attrbRefs.cbegin ();
            while (it != attrbRefs.cend () &&
                   it->attrbRef.num == NIOsymAttrbNum)
              ++it;

            if (it == attrbRefs.cend ())
              {
                // only NIO pseudo attribute occurrences were found, none of
                // which can be the dependent attribute; issue an error message
                // for the first occurrence

                const attrbPair& attref = attrbRefs.front ().attrbRef;

                const prodSymbol& symbol = prodBuffer2[attref.pos];

                yox_error (attrbRefs.front ().loc,
                           "\"" +
                               symbol.lexeme + "." +
                               std::to_string (symbol.symOccr) + "." +
                               NIOsym +
                           "\""
                           " cannot be a dependent.\n");
              }
            else
              {
                // register the dependent occurrence
                (void)memberDepend (it->attrbRef, true, true);
              }

            // remaining non-NIO attribute occurrences are the dependees
            while (++it != attrbRefs.cend ())
              {
                const attrbPair& attref = it->attrbRef;

                if (attref.num == NIOsymAttrbNum)
                  continue; // skip any NIO pseudo attribute occurrences

                if (!memberDepend (attref, true, false))
                  {
                    (void)memberDepend (attref, false, true);
                    continue;
                  }

                const prodSymbol& symbol = prodBuffer2[attref.pos];

                yox_error (it->loc,
                           "\"" +
                               symbol.lexeme + "." +
                               std::to_string (symbol.symOccr) + "." +
                               (*(gSymTable[symbol.lexeme]
                                      .symSort.attrbSeq))[attref.num].id +
                           "\""
                           " depends on itself.\n");
              }
          }
      }
  |   "@m" dependent_name_seq ";" attrb_evaluation
      {
        const attrbRefSeq& attrbRefs = $attrb_evaluation;

        for (const attrbRefSeqElem& it : attrbRefs)
          {
            const attrbPair& attref = it.attrbRef;

            if (attref.num == NIOsymAttrbNum)
              continue; // skip any NIO pseudo attribute occurrences

            if (!memberDepend (attref, true, false))
              (void)memberDepend (attref, false, true);
          }
      }
  |   "@e" dependent_name_seq ":" dependee_name_seq ";" attrb_evaluation
      {
        const attrbRefSeq& attrbRefs = $attrb_evaluation;

        for (const attrbRefSeqElem& it : attrbRefs)
          {
            const attrbPair& attref = it.attrbRef;

            if (attref.num == NIOsymAttrbNum)
              continue; // skip any NIO pseudo attribute occurrences

            // attribute is explicitly named as a dependent or a dependee
            if (memberDepend (attref, false, false)
                || memberDepend (attref, true, false))
              continue;

            // attribute is already known to be bogus
            if (attref.pos == -1 || attref.num == -1)
              continue;

            const prodSymbol& symbol = prodBuffer2[attref.pos];

            yox_error (it.loc,
                       "\"" +
                           symbol.lexeme + "." +
                           std::to_string (symbol.symOccr) + "." +
                           (*(gSymTable[symbol.lexeme]
                                  .symSort.attrbSeq))[attref.num].id +
                       "\""
                       " is not explicitly named.\n");
          }
      }
  ;

attrb_evaluation
  :   attrb_evaluation_stmt ";"
      {
        $attrb_evaluation = $attrb_evaluation_stmt;
      }
  |   attrb_evaluation_compound_stmt
      {
        $attrb_evaluation = $attrb_evaluation_compound_stmt;
      }
  ;

attrb_evaluation_stmt
  :   attrb_evaluation_expr
      {
        $attrb_evaluation_stmt = $attrb_evaluation_expr;
      }
  ;

attrb_evaluation_expr[left] // visible tokens in an expression, or statement

      // process a sequence of tokens legal in an expression, or in a
      // statement w/ expressions: identifier (includes C/C++ keyword),
      // literal, attribute reference, operator (includes ',') and
      // nested '()' & '[]' pairs

      // NOTE: better syntax errors requires C/C++ statement/expression syntax

  :   %empty
      {
        @left = yyla.location;
        $left = attrbRefSeq (); // empty sequence
      }
  |   attrb_evaluation_expr[right] attrb_evaluation_expr_component
      {
        attrbRefSeq new_seq = $right;

        const attrbRefSeq& component = $attrb_evaluation_expr_component;

        try
          {
            new_seq.insert (new_seq.end (),
                            component.cbegin (),
                            component.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  |   error
      {
        yyclearin;
        $left = attrbRefSeq (); // empty sequence
      }
  ;

attrb_evaluation_expr_component
  :   attrb_ref
      {
        const attrbPair& attrb_ref = $attrb_ref;

        rsBuffer.putAttrbPair (attrb_ref);

        attrbRefSeq new_seq = attrbRefSeq (); // empty sequence

        try
          {
            new_seq.push_front (attrbRefSeqElem (attrb_ref, @attrb_ref));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $attrb_evaluation_expr_component = new_seq;
      }
  |   expr_operator
      {
        $attrb_evaluation_expr_component = attrbRefSeq (); // empty sequence
      }
  |   expr_value
      {
        $attrb_evaluation_expr_component = attrbRefSeq (); // empty sequence
      }
  |   "(" attrb_evaluation_stmt_seq ")" // sub-expression, parameter list,
                                        // 'for ( ... )'
      {
        $attrb_evaluation_expr_component = $attrb_evaluation_stmt_seq;
      }
  |   "[" attrb_evaluation_expr "]"
      {
        $attrb_evaluation_expr_component = $attrb_evaluation_expr;
      }
  ;

attrb_ref
  :   OX_ATTRB_REF
      {
        $attrb_ref = transAttrbTripleToAttrbPair (@OX_ATTRB_REF, $OX_ATTRB_REF);
      }
  ;

attrb_evaluation_compound_stmt
  :   "{" attrb_evaluation_stmt_seq "}"
      {
  $attrb_evaluation_compound_stmt = $attrb_evaluation_stmt_seq;
      }
  ;

attrb_evaluation_stmt_seq[left] // visible tokens in a statement sequence

      // process a sequence of tokens legal in a statement sequence
      // w/ expressions: identifier (includes C/C++ keyword), literal,
      // attribute reference, operator (includes ','), nested '()' pair,
      // nested '{}' pair and ';'

      // NOTE: better syntax errors requires C/C++ statement/expression syntax

  :   %empty
      {
        @left = yyla.location;
        $left = attrbRefSeq (); // empty sequence
      }
  |   attrb_evaluation_stmt_seq[right] attrb_evaluation_stmt_seq_component
      {
        attrbRefSeq new_seq = $right;

        const attrbRefSeq& component = $attrb_evaluation_stmt_seq_component;

        try
          {
            new_seq.insert (new_seq.end (),
                            component.cbegin (),
                            component.cend ());
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        $left = new_seq;
      }
  |   error
      {
        yyclearin;
        $left = attrbRefSeq (); // empty sequence
      }
  ;

attrb_evaluation_stmt_seq_component
  :   attrb_evaluation_stmt_component
      {
        $attrb_evaluation_stmt_seq_component = $attrb_evaluation_stmt_component;
      }
  |   attrb_evaluation_compound_stmt
      {
        $attrb_evaluation_stmt_seq_component = $attrb_evaluation_compound_stmt;
      }
  |   /* empty */ ";"
      {
        $attrb_evaluation_stmt_seq_component = attrbRefSeq (); // empty sequence
      }
  ;

attrb_evaluation_stmt_component
  :   attrb_evaluation_expr_component
      {
        $attrb_evaluation_stmt_component = $attrb_evaluation_expr_component;
      }
  ;

// mixed and explicit mode

dependent_name_seq
  :   dependent_name
  |   dependent_name_seq dependent_name
  ;

dependent_name
  :   attrb_occur
      {
        const attrbPair& attref = $attrb_occur;

        if (attref.num == NIOsymAttrbNum)
          {
            // NIO pseudo attribute reference

            const prodSymbol& symbol = prodBuffer2[attref.pos];

            yox_error (@attrb_occur,
                       "\"" +
                           symbol.lexeme + "." +
                           std::to_string (symbol.symOccr) + "." +
                           NIOsym +
                       "\""
                       " cannot be a dependent.\n");
          }
        else if (memberDepend (attref, true, true))
          {
            const prodSymbol& symbol = prodBuffer2[attref.pos];

            yox_error (@attrb_occur,
                       "duplicate dependent "
                       "\"" +
                           symbol.lexeme + "." +
                           std::to_string (symbol.symOccr) + "." +
                           (*(gSymTable[symbol.lexeme]
                                  .symSort.attrbSeq))[attref.num].id +
                       "\".\n");
          }
      }
  |   error
      {
        yyclearin;
      }
  ;

attrb_occur
  :   OX_ATTRB_OCCUR
      {
        $attrb_occur = transAttrbTripleToAttrbPair (@OX_ATTRB_OCCUR,
                                                    $OX_ATTRB_OCCUR);
      }
  |   attrb_ref // hack: avoid a common syntax error
      {
        $attrb_occur = $attrb_ref;
      }
  ;

// explicit mode

dependee_name_seq
  :   %empty
  |   dependee_name dependee_name_seq
  |   error
      {
        yyclearin;
      }
  ;

dependee_name
  :   attrb_occur
      {
        const attrbPair& attref = $attrb_occur;

        if (attref.num == NIOsymAttrbNum)
          {
            // NIO pseudo attribute reference

            const prodSymbol& symbol = prodBuffer2[attref.pos];

            yox_error (@attrb_occur,
                       "\"" +
                           symbol.lexeme + "." +
                           std::to_string (symbol.symOccr) + "." +
                           NIOsym +
                       "\""
                       " cannot be a dependee.\n");
          }
        else if (memberDepend (attref, true, false))
          {
            const prodSymbol& symbol = prodBuffer2[attref.pos];

            yox_error (@attrb_occur,
                       "\"" +
                           symbol.lexeme + "." +
                           std::to_string (symbol.symOccr) + "." +
                           (*(gSymTable[symbol.lexeme]
                                  .symSort.attrbSeq))[attref.num].id +
                       "\""
                       " depends immediately on itself.\n");
          }
        else if (memberDepend (attref, false, true))
          {
            const prodSymbol& symbol = prodBuffer2[attref.pos];

            yox_error (@attrb_occur,
                       "duplicate dependee "
                       "\"" +
                           symbol.lexeme + "." +
                           std::to_string (symbol.symOccr) + "." +
                           (*(gSymTable[symbol.lexeme]
                                  .symSort.attrbSeq))[attref.num] .id +
                       "\".\n");
          }
      }
  ;

// traversal specification

traversal_segment
  :   "traversal mode annunciator"[OX_ID]
      {
        ++dynTravModSeq;

        try
          {
            dynTravModArr.push_back (dynTravMod ($OX_ID));
          }
        catch (...)
          {
            yox_alloc_fail (__FILE__, __LINE__);
          }

        // throw away any pending comments & white space

        startTravAction (@OX_ID); // overwritten if next is a traversal modifier
      }
      traversal_modifier_seq
      traversal_action
      {
        wrapUpTravAction ();
      }
  ;

traversal_modifier_seq[left]
  :   %empty
      {
        @left = yyla.location;
      }
  |   traversal_modifier_seq[right] traversal_modifier
  ;

traversal_modifier
  :   "@revorder"[REVORDER]
      {
        // @revorder may occur once per traversal action

        if (++dynTravModArr[dynTravModSeq].orderRev > 1)
          yox_error (@REVORDER,
                     "multiple @revorder modifiers for traversal "
                     "\"" + dynTravModArr[dynTravModSeq].id + "\".\n");

        // throw away any pending comments & white space

        startTravCondition (@REVORDER);
      }
      traversal_modifier_expr
      {
        wrapUpTravOrderCond ();

        // throw away any pending comments & white space

        startTravAction (@traversal_modifier_expr);
        // overwritten if next is a traversal modifier
      }
  |   "@revdirection"[REVDIRECTION]
      {
        // @revdirection may occur once per traversal action

        if (++dynTravModArr[dynTravModSeq].dirRev == 1)
          dynTravModArr[dynTravModSeq].dirRevLoc = @REVDIRECTION;
        else if (dynTravModArr[dynTravModSeq].id != errTravID)
          yox_error (@REVDIRECTION,
                     "multiple @revdirection modifiers for traversal "
                     "\"" + dynTravModArr[dynTravModSeq].id + "\".\n");

        // throw away any pending comments & white space

        startTravCondition (@REVDIRECTION);
      }
      traversal_modifier_expr
      {
        wrapUpTravDirCond ();

        // throw away any pending comments & white space

        startTravAction (@traversal_modifier_expr);
        // overwritten if next is a traversal modifier
      }
  ;

traversal_modifier_expr
  :   "(" traversal_expr ")"
  ;

traversal_expr[left] // visible tokens in an expression, or statement

      // process a sequence of tokens legal in an expression, or in a
      // statement w/ expressions: identifier (includes C/C++ keyword),
      // literal, attribute reference, operator (includes ',') and
      // nested '()' & '[]' pairs

      // NOTE: better syntax errors requires C/C++ statement/expression syntax

  :   %empty
      {
        @left = yyla.location;
      }
  |   traversal_expr[right] traversal_expr_component
  |   error
      {
        yyclearin;
      }
  ;

traversal_expr_component
  :   attrb_ref
      {
        rsBuffer.putAttrbPair ($attrb_ref);
      }
  |   expr_operator
  |   expr_value
  |   "(" traversal_expr ")" // sub-expression, parameter list, 'for ( ... )'
  |   "[" traversal_expr "]"
  ;

traversal_action
  :   traversal_stmt ";"
  |   traversal_compound_stmt
  ;

traversal_stmt
  :   traversal_expr
  ;

traversal_stmt_component
  :   traversal_expr_component
  ;

traversal_compound_stmt
  :   "{" traversal_stmt_seq "}"
  ;

traversal_stmt_seq[left] // visible tokens in a statement sequence

      // process a sequence of tokens legal in a statement sequence
      // w/ expressions: identifier (includes C/C++ keyword), literal,
      // attribute reference, operator (includes ','), nested '()' pair,
      // nested '{}' pair and ';'

      // NOTE: better syntax errors requires C/C++ statement/expression syntax

  :   %empty
      {
        @left = yyla.location;
      }
  |   traversal_stmt_seq[right] traversal_stmt_seq_component
  ;

traversal_stmt_seq_component
  :   traversal_stmt_component
  |   traversal_compound_stmt
  |   ";"
  ;

expr_value
  :   "identifier"
  |   LITERAL_CHAR
  |   LITERAL_STRING
  |   "number"
  ;

expr_operator // C/C++ operators
  :   "."
  |   "->"
  |   "++"
  |   "--"
  |   "~"
  |   "!"
  |   "-"
  |   "+"
  |   "&"
  |   "*"
  |   "/"
  |   "%"
  |   "<<"
  |   ">>"
  |   "<"
  |   ">"
  |   "<="
  |   ">="
  |   "=="
  |   "!="
  |   "^"
  |   "|"
  |   "&&"
  |   "||"
  |   "?"
  |   ":"
  |   "="
  |   "+="
  |   "-="
  |   "*="
  |   "/="
  |   "%="
  |   "<<="
  |   ">>="
  |   "&="
  |   "^="
  |   "|="
  |   ","
  |   "->*"
  |   ".*"
  |   "::"
  ;

%%

// called from 'main ()'
extern "C" void
agc_bison_parse (
#if defined(WITH_REFLEX)
                          const reflex::Input& yoxIn,
#else // defined(WITH_FLEX)
                          std::ifstream& yoxIn,
#endif
                          Yfile_specIO& Yfile,
                          int debug)
{
  AGC::bisonLexer agc_Yfile_lexer (Yfile, yoxIn);
  agc_Yfile_lexer.set_debug (debug >> 1);
  AGC::bison parser (agc_Yfile_lexer, outMgr::ctrl, Yfile);
  parser.set_debug_level (debug & 0x1);
  (void)parser ();
}

void
AGC::bison::error (const AGC::location& locp, const std::string& msg)
{
  agc_syntax_error (locp, msg);
}

void
AGC::bison::report_syntax_error (const context& yyctx) const
{
  // # of expected tokens
  enum
  {
    YYARGS_MAX = 4
  };

  symbol_kind_type yyarg[YYARGS_MAX] = { symbol_kind::S_YYEMPTY };

  // Actual size of yyarg[]
  int yycount = yyctx.expected_tokens (yyarg, YYARGS_MAX);

  if (yycount < 0)
    return;
  if ((yycount == 0) && (yyarg[0] != (symbol_kind_type)symbol_kind::S_YYEMPTY))
    yycount = YYARGS_MAX;

  const std::string token_name = std::string (symbol_name (yyctx.token ()));

  std::ostringstream yymsg;

  yymsg << "syntax error";

  if (yycount > 0)
    {
      yymsg << ", unexpected ";

      switch (yyctx.token ())
        {
        case symbol_kind::S_ID:             // identifier
        case symbol_kind::S_GS_ID:          // Yacc symbol name
        case symbol_kind::S_OX_ID:          // traversal mode annunciator
        case symbol_kind::S_NAMESPACE_ID:   // C++ qualified name
        case symbol_kind::S_INHATTR_ID:     // inherited attribute
        case symbol_kind::S_API_VALUE:      // API value
        case symbol_kind::S_API_VARIABLE:   // API variable
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<name_literal> () +
                       "]";
          break;

        case symbol_kind::S_LITERAL_CHAR:
        case symbol_kind::S_LITERAL_STRING:
        case symbol_kind::S_LITERAL_PATH:   // literal path name
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<char_literal> () +
                       "]";
          break;

        case symbol_kind::S_NUM:            // number
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<numberRep> ().spelling +
                       "]";
          break;

        case symbol_kind::S_OX_ATTRB_REF:   // attribute reference
        case symbol_kind::S_OX_ATTRB_OCCUR: // attribute occurrence
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<attrbTriple> ().gsName +
                       "." +
                       std::to_string (yyctx.lookahead ()
                                           .value.as<attrbTriple> ()
                                           .occurrence) +
                       "." +
                       yyctx.lookahead ().value.as<attrbTriple> ().atName +
                       "]";
          break;

        case symbol_kind::S_CODE:           // %code, %initial-action
        case symbol_kind::S_DESTRUCTOR:     // %destructor
        case symbol_kind::S_EXPECT:         // %expect
        case symbol_kind::S_EXPECT_RR:      // %expect_rr
        case symbol_kind::S_LEX_PARAM:      // %lex-param
        case symbol_kind::S_PARSE_PARAM:    // %parse-param
        case symbol_kind::S_PURE_PARSER:    // %pure-parser
        case symbol_kind::S_DEFINE:         // %define
        case symbol_kind::S_DPREC:          // %dprec
        case symbol_kind::S_EMPTY:          // %empty
        case symbol_kind::S_LANGUAGE:       // %language
        case symbol_kind::S_MERGE:          // %merge
        case symbol_kind::S_NAME_PREFIX:    // %name-prefix
        case symbol_kind::S_NTERM:          // %nterm
        case symbol_kind::S_PARAM:          // %param
        case symbol_kind::S_PRECEDENCE:     // %precedence
        case symbol_kind::S_PRINTER:        // %printer
        case symbol_kind::S_SKELETON:       // %skeleton
        case symbol_kind::S_LOCATION:       // %location, %position, %@
        case symbol_kind::S_EXPORT:         // %export
        case symbol_kind::S_IMPORT:         // %import
        case symbol_kind::S_LOCAL:          // %local
        case symbol_kind::S_LOOKAHEAD:      // %lookahead
          yymsg << "\'" + token_name + "\'";
          break;

        default:
          yymsg << "\'" + token_name + "\'";
          break;
        }

      std::string trailing_list;
      for (int i = yycount; --i > 0;)
        {
          std::string temp;

          if (i + 1 == yycount)
            temp = " or";
          else
            temp = ",";

          trailing_list =
              temp + " \'" + symbol_name (yyarg[i]) + "\'" + trailing_list;
        }

      yymsg << ", expecting \'" << symbol_name (yyarg[0]) << "\'"
            << trailing_list;
    }
  yymsg << ".";

  AGC::location loc = yyctx.location ();
  agc_syntax_error (loc, yymsg);
}
