%start GEN YACC_TRAILER
%start LEX_DEFS RULES_PROLOG
%start LEX_REGEXP LEX_ACT GEN_ACT
%start LEX_CCODE CCODE_BLOCK LEX_TRAILER
%start COMMENT SKIP_COMMENT
%start STRING SKIP_STRING

%start MACRO_DEF
%start ATTR_DECL ATTR_DECL_SYMS
%start TRAV_DECL
%start YATTR_REF LATTR_REF

%option c++

%option never-interactive stack

%option noyywrap
%option noyy_top_state
%option nounput

%{
/* $Id: agc-strip-lexer.ll 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <string>

#include "YY_LOCATIONS.hh"
%}

%top{
#include <fstream>

#include "agc-Lfile.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"

#define YY_DECL \
  int AGC::STRIP::Lexer::yylex (AGC::lexer::value_type& yylval \
                                    __attribute__ ((unused)), \
                                AGC::location& yylloc)

// inhibit the 'flex' builtin definition of 'struct yy_buffer_state'
#define YY_STRUCT_YY_BUFFER_STATE
}

%{
#include "agc-strip-lexer.hh"
%}

WS		[ \t\f\v]+
OPTWS		[ \t\f\v]*
NON_WS		[^ \t\f\v\n]

ESCSEQ		\\([^\n]|[0-7]{1,3}|x[[:xdigit:]]{1,2})
CHAR_LIT	"'"([^\\'\n]|{ESCSEQ})"'"
STRING_LIT	"\""([^\\"\n]|{ESCSEQ})*"\""

CHAR_CONST	"L"?{CHAR_LIT}
STRING_CONST	"L"?{STRING_LIT}

ID		[a-zA-Z_][a-zA-Z_0-9]*
DOT_NAME	[_\.]*[a-zA-Z][-_\.a-zA-Z0-9]*

INTEGER		[0-9]+
HEX		"0"[xX][0-9a-fA-F]+
OCTAL		"0"[0-7]+
INT_SUFFIX	([uU][lL]{0,2})|([lL]{1,2}[uU]?)
NUMBER		({INTEGER}|{HEX}|{OCTAL}){INT_SUFFIX}?

SYMBOL		{DOT_NAME}|{CHAR_LIT}

START_CONDS	"<"{ID}(","{ID})*">"

PAT_NAME	[_a-zA-Z][-_a-zA-Z0-9]*
NAMED_PAT	"{"{PAT_NAME}"}"

DIFF_OR_UNION	"{-}"|"{+}"

CHARCLASS_PAT	"["([^\\\]\n]|{ESCSEQ})+"]"
OTHER		([^<*+?{}^$" [\]\t\f\v\n][^*+?{}^$" [\]\t\f\v\n]?)|{DIFF_OR_UNION}
REPEAT		\*|\+|\?|"{"{INTEGER}(","{INTEGER}?)?"}"
L_ANCHOR	"^"
R_ANCHOR	"$"
COMPONENT	({STRING_LIT}|{ESCSEQ}|{CHARCLASS_PAT}|{NAMED_PAT}|{OTHER}){REPEAT}?
PATTERN		{L_ANCHOR}?{COMPONENT}+{R_ANCHOR}?

%%

<GEN>{

  /* skip Ox-specific content in the Y-file */

"/*" {
  yoxOutG.put (text ());
  yox_push_state (COMMENT, yoxOutG);
}

"L"?"\"" {
  yoxOutG.put (text ());
  yox_push_state (STRING, yoxOutG);
}

"@attributes" {
  push_state (yylloc, ATTR_DECL);
}

"@macro" {
  push_state (yylloc, MACRO_DEF);
}

"@autoinh"|"@autosyn"|"@traversal" {
  push_state (yylloc, TRAV_DECL);
}

"@{" {
  push_state (yylloc, YATTR_REF);
}

^{OPTWS}"%%".*$ {
  if (++yfileMarkSeen == 2)
    start (YACC_TRAILER);
  yoxOutG.put (text ());
}

"//".*$ {
  yoxOutG.put (text ());
}

\n|. {
  yoxOutG.put ((char)chr ());
}

}



<MACRO_DEF>{

  /* GEN 'subroutine': skip Ox macro definition */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutG);
}

"L"?"\"" {
  yox_push_state (SKIP_STRING, yoxOutG);
}

"@end" {
  pop_state ();
}

\n {
  yoxOutG.put ((char)chr ());
}

"//".*$|. {
  ; // no-op
}

}



<YATTR_REF>{

  /* GEN 'subroutine': Ox Y-file attribute reference section */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutG);
}

"L"?"\"" {
  yox_push_state (SKIP_STRING, yoxOutG);
}

"@}" {
  pop_state ();
}

"@{" {
  unexpectedATLBRACE (yylloc);
}

\n {
  yoxOutG.put ((char)chr ());
}

"//".*$|. {
  ; // no-op
}

}



<ATTR_DECL>{

  /* GEN 'subroutine': skip '@attributes' declaration; the closing brace
     transitions to ATTR_DECL_SYMS, which returns */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutG);
}

"}" {
  start (ATTR_DECL_SYMS);
}

\n {
  yoxOutG.put ((char)chr ());
}

"//".*$|{CHAR_CONST}|. {
  ; // no-op
}

<<EOF>> {
  unterminatedAttrDeclaration (yylloc, startPos.top ());
}

}



<ATTR_DECL_SYMS>{

  /* GEN 'subroutine': skip grammar symbols in an attribute declaration
     symbol list */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutG);
}

"//".*$ {
  yoxOutG.put (text ());
}

{SYMBOL}|{WS} {
  ; // no-op
}

\n {
  yoxOutG.put ((char)chr ());
}

. {
  yyless (0);
  pop_state ();
}

}



<TRAV_DECL>{

  /* GEN 'subroutine': skip Ox traversal declaration */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutG);
}

"//".*$ {
  yoxOutG.put (text ());
}

"@postorder"|"@preorder" |
"@lefttoright"|"@righttoleft" |
"@disable" |
"@warn" |
{ID}|{WS} {
  ; // no-op
}

\n {
  yoxOutG.put ((char)chr ());
}

. {
  yyless (0);
  pop_state ();
}

}



<LEX_DEFS>{

  /* L-file Definitions Section */

^"%{".*$ {
  yoxOutL.put (text ());
  push_state (yylloc, LEX_CCODE);
}

"%top" {
  yoxOutL.put (text ());
  push_state (yylloc, CCODE_BLOCK);
  nLbraces = 0;
}

"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

^{OPTWS}"%%".*$ {
  start (RULES_PROLOG);
  yoxOutL.put (text ());
}

"%option".*$ {
  yoxOutL.put (text ());
}

"//".*$ {
  yoxOutL.put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

}



<LEX_CCODE>{

  /* LEX_DEFS, LEX_REGEXP 'subroutine': echo C code between "%{" and "%}" */

"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.put (text ());
  yox_push_state (STRING, yoxOutL);
}

^"%}".*$ {
  yoxOutL.put (text ());
  pop_state ();
}

"//".*$ {
  yoxOutL.put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<CCODE_BLOCK>{

  /* LEX_DEFS 'subroutine': echo C code "{ ... }" for 'flex' extension "%top" */

"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.put (text ());
  yox_push_state (STRING, yoxOutL);
}

"{" {
  yoxOutL.put ((char)chr ());
  nLbraces++;
}

"}" {
  yoxOutL.put ((char)chr ());
  if (--nLbraces == 0)
    pop_state ();
}

"//".*$|{CHAR_CONST} {
  yoxOutL.put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<RULES_PROLOG>{

  /* echo the prolog of the Lex Rules Section */

^{NON_WS} {
  yyless (0);
  start (LEX_REGEXP);
  nLbraces = 0;
}

\n|. {
  yoxOutL.put ((char)chr ());
}

}



<LEX_REGEXP>{

  /* echo Lex regular expressions, terminated by whitespace */

^{OPTWS}"%%".*$ {
  yoxOutL.put (text ());
  start (LEX_TRAILER);
}

^"%{".*$ {
  yoxOutL.put (text ());
  push_state (yylloc, LEX_CCODE);
}

^{OPTWS}"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

{WS} {
  yoxOutL.put (text ());
  push_state (yylloc, LEX_ACT);
}

^{OPTWS}"//".*$ {
  yoxOutL.put (text ());
}

{WS}"|".*$ {
  yoxOutL.put (text ());
}

{WS}$ {
  yoxOutL.put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

}



<GEN_ACT,LEX_ACT>{

  /* GEN_ACT: initial start condition if "theFiles.scanner == GEN_SPEC_LEXER" */

  /* LEX_ACT: LEX_REGEXP 'subroutine' */
  /* look for the start of an L-file attribute reference section (which sends
     this lexer to LATTR_REF), and look for the end of a Lex action after an
     attribute reference section; a Lex action is terminated by the first
     newline outside of paired braces */

"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.put (text ());
  yox_push_state (STRING, yoxOutL);
}

"@{" {
  push_state (yylloc, LATTR_REF);
}

"{" {
  yoxOutL.put ((char)chr ());
  nLbraces++;
}

"}" {
  yoxOutL.put ((char)chr ());
  nLbraces--;
}

"//".*$|{CHAR_CONST} {
  yoxOutL.put (text ());
}

\n {
  yoxOutL.put ((char)chr ());
  if ((start () == LEX_ACT) && (nLbraces == 0))
    pop_state ();
}

. {
  yoxOutL.put ((char)chr ());
}

<<EOF>> {
  if (start () == LEX_ACT)
    unterminatedAction (yylloc, startPos.top ());

  return lexer::token::END;
}

}



<LATTR_REF>{

  /* LEX_ACT 'subroutine': skip Ox L-file attribute reference section */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutL);
}

"L"?"\"" {
  yox_push_state (SKIP_STRING, yoxOutL);
}

"@}" {
  pop_state ();
}

"@{" {
  unexpectedATLBRACE (yylloc);
}

\n {
  yoxOutL.put ((char)chr ());
}

"//".*$|. {
  ; // no-op
}

<<EOF>> {
  unterminatedAttrRefSection (yylloc, startPos.top ());
}

}



<YACC_TRAILER>{

  /* echo the optional user section following the second "%%" in the Y-file,
     if present */

"/*" {
  yoxOutG.put (text ());
  yox_push_state (COMMENT, yoxOutG);
}

"L"?"\"" {
  yoxOutG.put (text ());
  yox_push_state (STRING, yoxOutG);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutG.put (text ());
}

\n|. {
  yoxOutG.put ((char)chr ());
}

}



<LEX_TRAILER>{

  /* echo the optional user section following the second "%%" in the L-file,
     if present */

"/*" {
  yoxOutL.put (text ());
  yox_push_state (COMMENT, yoxOutL);
}

"L"?"\"" {
  yoxOutL.put (text ());
  yox_push_state (STRING, yoxOutL);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutL.put (text ());
}

\n|. {
  yoxOutL.put ((char)chr ());
}

}



<COMMENT>{

  /* 'subroutine': handle multi-line C comments */

"*/" {
  whereTo->put (text ());
  pop_state ();
}

[^\*\n]+ {
  whereTo->put (text ());
}

\n|. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedComment (yylloc, startPos.top ());
}

}



<SKIP_COMMENT>{

  /* 'subroutine': skip multi-line C comments */

"*/" {
  pop_state ();
}

[^\*\n]+|\n|. {
  ; // no-op
}

<<EOF>> {
  unterminatedComment (yylloc, startPos.top ());
}

}



<STRING>{

  /* 'subroutine': handle multi-lint C/C++ string literals */

"\"" {
  whereTo->put ((char)chr ());
  if (stringHasBareNewline)
    stringContainsNewline (yylloc, startPos.top ());
  pop_state ();
}

[^"\\\n]+|\\(.|\n) {
  whereTo->put (text ());
}

\n {
  stringHasBareNewline = true;
  whereTo->put ((char)chr ());
}

. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedString (yylloc, startPos.top ());
}

}



<SKIP_STRING>{

  /* 'subroutine': skip multi-lint C/C++ string literals */

"\"" {
  if (stringHasBareNewline)
    stringContainsNewline (yylloc, startPos.top ());
  pop_state ();
}

\n {
  stringHasBareNewline = true;
}

[^"\\\n]+|\\(.|\n)|. {
  ; // no-op
}

<<EOF>> {
  unterminatedString (yylloc, startPos.top ());
}

}



<<EOF>> {
  return lexer::token::END;
}

%%

// strip lexer constructor
AGC::STRIP::Lexer::Lexer (specIO& fR, std::ifstream& yoxIn)
    : theFiles (fR),
      EOLstate (EOLctx (position (&fR.inName, 0, 0)))
{
  yycolumn = 1;

  nLbraces = 0;

  yfileMarkSeen = 0;

  yyrestart (yoxIn);

  if (theFiles.scanner == YACC_SPEC_LEXER)
    {
      start (GEN);
      return;
    }

  switch (theFiles.scanner)
    {
    case LEX_SPEC_LEXER:
      start (LEX_DEFS);
      break;
    case GEN_SPEC_LEXER:
      start (GEN_ACT);
      break;
    default:
      break;
    }
}

namespace AGC
{
namespace STRIP
{
#if defined(_WIN32)
#include "YY_INPUT.hh"
#endif

void
Lexer::yox_push_state (int lexState, outMgr& whTo)
{
  whereTo = &whTo;
  if (lexState == STRING || lexState == SKIP_STRING)
    stringHasBareNewline = false;
  push_state (*whereTo->getCurPos (), lexState);
}

}
}
