/* $Id: macutil.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef MACUTIL_HH_INCLUDED
#define MACUTIL_HH_INCLUDED

// during macro expansion, manage input buffers as a stack:
// maintain line, column and the EOL context in sync with the
// position in the input file and macro body (& arguments)

#define pushbuffer(newbuf, newposition) \
{ \
  /* save the current EOL state */ \
  YY_CURRENT_BUFFER_LVALUE->yy_bs_lineno = yylineno; \
  YY_CURRENT_BUFFER_LVALUE->yy_bs_column = yycolumn; \
  YY_CURRENT_BUFFER_LVALUE->yy_bs_loc = yylloc; \
  YY_CURRENT_BUFFER_LVALUE->yy_bs_EOLstate = EOLstate; \
  /* make newbuf the current buffer */ \
  yypush_buffer_state (newbuf); \
  yylloc = AGC::location (newposition); \
  yylineno = newposition.line; \
  yycolumn = newposition.column; \
  if (yycolumn == 1) \
    yy_set_bol (true); \
  EOLstate.atEOL = false; \
}

#define popbuffer \
{ \
  yypop_buffer_state (); \
  /* restore prior EOL state */ \
  yylineno = YY_CURRENT_BUFFER_LVALUE->yy_bs_lineno; \
  yycolumn = YY_CURRENT_BUFFER_LVALUE->yy_bs_column; \
  yylloc = YY_CURRENT_BUFFER_LVALUE->yy_bs_loc; \
  EOLstate = YY_CURRENT_BUFFER_LVALUE->yy_bs_EOLstate; \
}

#endif // !MACUTIL_HH_INCLUDED
