/* $Id: memMgr.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef MEMMGR_HH_INCLUDED
#define MEMMGR_HH_INCLUDED

#include <sstream>
#include <string>

#include "codegen.hh"
#include "outMgr.hh"



class fileMgr;
class lookaheadBuf;



class memMgr : public codeGen
{
public:

  memMgr ()
      : codeGen (new std::stringstream ())
  {
    wasOutName = false;
    origin = AGC::position (new std::string (), 0, 0);
    curPos = &origin;
  }

  memMgr (std::string name)
      : codeGen (new std::stringstream (), name, -1)
  {
    wasOutName = false;
    origin = AGC::position (&name, 0, 0);
    curPos = &origin;
  }

  memMgr (bool forLexer)
      : codeGen (new std::stringstream (),
                 new AGC::position (new std::string (), 0, 0),
                 forLexer)
  {
    wasOutName = false;
    curPos = &origin;
  }

  memMgr (std::string name, bool forLexer)
      : codeGen (new std::stringstream (),
                 name,
                 -1,
                 new AGC::position (&name, 0, 0),
                 forLexer)

  {
    wasOutName = false;
    curPos = &origin;
  }

  void clear (void);

  memMgr& put (char);

  inline memMgr&
  put (const std::string& str)
  {
    return static_cast<memMgr&> (outMgr::put (str));
  }

  inline memMgr&
  put (int i, int width = 0)
  {
    return static_cast<memMgr&> (outMgr::put (i, width));
  }

  inline memMgr&
  put (long l, int width = 0)
  {
    return static_cast<memMgr&> (outMgr::put (l, width));
  }

  inline memMgr&
  append (fileMgr&)
  {
    // no-op (should never actually be called)
    return *this;
  }

  memMgr& append (lookaheadBuf&);

protected:

  // get the internal memory file
  inline std::stringstream&
  getObj (void)
  {
    return static_cast<std::stringstream&> (*theObj);
  }

  inline long
  tellp (void)
  {
    return getObj ().tellp ();
  }

private:
};



class lookaheadBuf : public memMgr
{
public:

  lookaheadBuf ()
      : memMgr ()
  {}

  lookaheadBuf (std::string name)
      : memMgr (name)
  {}

  lookaheadBuf (bool forLexer)
      : memMgr (forLexer)
  {}

  lookaheadBuf (std::string name, bool forLexer)
      : memMgr (name, forLexer)
  {}

protected:

  friend class fileMgr;

private:
};



struct attrbRefSeg
{
  AGC::position origin;
  std::string buffer;

  attrbRefSeg ()
      : origin (AGC::position (new std::string (), 0, 0)),
        buffer (std::string ())
  {}

  attrbRefSeg (AGC::position start, std::string ref)
      : origin (start),
        buffer (ref)
  {}

  void
  clear (void)
  {
    origin = AGC::position (new std::string (), 0, 0);
    buffer.clear ();
  }
};

class refSegBuf : public memMgr
{
public:

  refSegBuf ()
      : memMgr ()
  {}

  void init (const AGC::position);

  void putAttrbPair (attrbPair);

  inline attrbRefSeg
  get (void)
  {
    flush (); // write any currently buffered text to the internal memory file
    return attrbRefSeg (origin, getObj ().str ());
  }

protected:

  void putOutLineDirective (void); // never actually generate a line# directive

private:
};

#endif // !MEMMGR_HH_INCLUDED
