/* $Id: outMgr.hh 2837 2022-12-11 19:12:25Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2010-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef OUTMGR_HH_INCLUDED
#define OUTMGR_HH_INCLUDED

#include <fstream>
#include <string>

#include "opts.hh"



class fileMgr;
class lookaheadBuf;
struct attrbRefSeg;



class outMgr
{
public:

  static yoxControl ctrl;

  outMgr (std::iostream* obj,
          AGC::position* pos = nullptr,
          bool forLexer = false)
      : theObj (obj),
        name (std::string ()),
        outLineno (1),
        atBOL (true),
        lastInlineno (-1),
        whitespace (std::string ()),
        blankLines (0),
        curPos (pos),
        lexerOutput (forLexer)
  {
    if (curPos == nullptr)
      {
        origin = AGC::position (&name);
        curPos = &origin;
      }
    else
      origin = *curPos;
  }

  outMgr (std::iostream* obj,
          std::string name,
          int line,
          AGC::position* pos = nullptr,
          bool forLexer = false)
      : theObj (obj),
        name (name),
        outLineno (1),
        atBOL (true),
        lastInlineno (line),
        whitespace (std::string ()),
        blankLines (0),
        origin (*pos),
        curPos (pos),
        lexerOutput (forLexer)
  {
    if (curPos == nullptr)
      {
        origin = AGC::position (&name);
        curPos = &origin;
      }
    else
      origin = *curPos;
  }

  inline void
  setOrigin (const AGC::position& pos)
  {
    origin = pos;
  }

  inline void
  setCurPos (const AGC::position& pos)
  {
    curPos = &pos;
  }

  inline const AGC::position&
  getOrigin (void)
  {
    return origin;
  }

  inline const AGC::position*
  getCurPos (void)
  {
    return curPos;
  }

  inline const std::string&
  getName (void)
  {
    return name;
  }

  void clear (void); // clear the output object

  bool empty (void); // is the output object empty?

  virtual outMgr& put (char) = 0;

  virtual outMgr& put (const std::string&);

  virtual outMgr& put (int, int width = 0);

  virtual outMgr& put (long, int width = 0);

  virtual outMgr& lex (void);

  virtual outMgr& gen (void);

  virtual outMgr& append (fileMgr&) = 0;

  virtual outMgr& append (lookaheadBuf&) = 0;

protected:

  std::iostream* theObj; // the output object

  std::string name; // name of this output object

  bool wasOutName; // =true if most recent `#line' directive was "name"

  int outLineno; // line# of the output line currently under construction

  bool atBOL; // help generate #line directives correctly

  int lastInlineno; // line# of the last input copied to the output object

  std::string whitespace; // buffer whitespace, remove if precedes EOL

  std::string buffer; // hold the output line currently under construction

  int blankLines; // count of consecutive blank lines

  AGC::position origin; // starting position of this output object

  AGC::position const* curPos; // current position in this output object

  virtual std::iostream& getObj (void) = 0; // get the external file

  void putCharInWriteBuffer (char);

  bool flushBlankLines (void);

  void flush (void);

  virtual long tellp (void) = 0;

  bool lexerOutput; // =true if this object is for Lexer output

  std::string lineDirective (int, const std::string&, bool);

  void putLineDirective (const int, const std::string&, bool);

  virtual void putOutLineDirective (void);

  void putInLineDirective (int, const std::string&);

  void append (std::iostream&, std::string&);

private:
};

#endif // !OUTMGR_HH_INCLUDED
