/* $Id: trav.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef TRAV_HH_INCLUDED
#define TRAV_HH_INCLUDED

#include "location.hh"

#include "outMgr.hh"
#include "scan.hh"

extern bool travSpecConflict (void);
extern void newTrav (const AGC::location&, const std::string&);
extern void clearTravBuf (void);

extern std::string errTravID;
extern void clearErrorTrav (void);

extern void prepareDataStructures2 (void);

inline void
startTravCondition (const AGC::location& loc)
{
  rsBuffer.init (loc.begin);
}

inline void
startTravAction (const AGC::location& loc)
{
  rsBuffer.init (loc.end);
}

extern void wrapUpTravAction (void);
extern void wrapUpTravDirCond (void);
extern void wrapUpTravOrderCond (void);

extern int travLookup (const std::string&);

extern void travSetPreorder (void);
extern void travSetPostorder (void);
extern void travSetLeftToRight (void);
extern void travSetRightToLeft (void);
extern void travSetDisable (void);

extern bool travDisabled (void);

extern int currentTrav;

#endif // !TRAV_HH_INCLUDED
