/* Infix notation calculator--calc */
%define api.pure %define api.push_pull "both"

%code top {
#include "types.h"
}

%code provides /* -> header file, following the definition of YYSTYPE */
{
  /* the expected prototype of yylex:
   *   1st argument must be named `yylval' (because %define api.pure)
   */
#define YY_DECL \
  enum yytokentype yylex (YYSTYPE *yylval)
  YY_DECL;
}

%code
{
#include <assert.h>

static int power (int base, int exponent);

#include <stdio.h>

static void yyerror (const char *msg);
}

/* Grammar Declarations */
%token ERROR
%token NUM

%nonassoc '='   /* comparison            */
%left '-' '+'
%left '*' '/'
%nonassoc NEG   /* negation--unary minus */
%right '^'      /* exponentiation        */

@attributes {semantic_error eVal;} ERROR
@attributes {semantic_value iVal;} exp NUM
@attributes {int print;} line

%%
input	:	line
	|	input line
	;

line	:	'\n'		@{ @i @line.print@ = 0; @}
	|	exp '\n'	@{ @decorate
				   @i @line.print@ = fprintf(stderr, "semval = %d\n", @exp.iVal@);
				@}
	;

exp	:	NUM		@{ @i @exp.iVal@ = @NUM.iVal@; @}
	|	exp '=' exp	@{ @i @exp.0.iVal@ = (@exp.1.iVal@ == @exp.2.iVal@); @}
	|	exp '+' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ + @exp.2.iVal@; @}
	|	exp '-' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ - @exp.2.iVal@; @}
	|	exp '*' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ * @exp.2.iVal@; @}
	|	exp '/' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ / @exp.2.iVal@; @}
	|	'-' exp %prec NEG @{ @i @exp.0.iVal@ = -@exp.1.iVal@; @}
	|	exp '^' exp	@{ @i @exp.0.iVal@ = power(@exp.1.iVal@, @exp.2.iVal@); @}
	|	'(' exp ')'	@{ @i @exp.0.iVal@ = @exp.1.iVal@; @}
	|	ERROR		@{ @e exp.iVal : ERROR.eVal ;
				   {@exp.iVal@ = 1111;
				    fprintf(stderr, "gc: invalid input: [%c]\n", @ERROR.eVal@);}
				@}
	;
%%

static int
power (int base, int exponent)
{
  int res = 1;
  assert (0 <= exponent);
  for (/* Niente */; exponent; --exponent)
    res *= base;
  return res;
}

static
void yyerror (const char *msg)
{
  fprintf(stderr, "%s\n", msg);
}
