%top {
// -> header file, at the top

// ensure that "system" includes precede
// includes in the lexer skeleton

#include <cerrno>
#include <climits>
#include <cstdlib>
#include <string>

// the `bison' defined `yylex' prototype:
//   1st argument must be named `yylval_param' 
//   2nd argument must be named `yylloc_param'
//   3rd argument must be named `drv'

#define YYLEX_ARGS \
  YYSTYPE* yylval, YYLTYPE* yylloc, driver& drv

#if defined (USE_FLEXCXX)
#define YY_DECL \
  int yoxLexer::yylex (YYLEX_ARGS)
#else
#define YY_DECL \
  int           yylex (YYLEX_ARGS)
#endif

}

%{
#include "parser.hh"
#include "driver.hh"

#if defined (USE_FLEXCXX)

class yoxLexer: public yyFlexLexer {
public:

 ~yoxLexer () {};

  yoxLexer (std::istream* arg_yyin = 0, std::ostream* arg_yyout = 0) :
    yyFlexLexer (arg_yyin, arg_yyout) {}

  int yylex (YYLEX_ARGS);

};

// begin hack: required by broken, IMHO, `flex' C++ scanner implememtation

// There is no way to override the signature of `yylex()' in the base
// class `FlexLexer' and `yyFlexLexer'.  The `YY_DECL' mechanism only
// applies to the application lexer class, derived from `yyFlexLexer'.

// A `bison' reentrant parser passed `yylval' (and `yylloc') as pointers
// to `yylex()' (a `bison' C++ parser is reentrant).  In order to enable
// the `flex' C++ lexer to access those values, `YY_DECL' is defined as
// the application lexer class' `yylex()' member function, leaving the
// `yyFlexLexer::yylex(void)' member function hanging out in the cold;
// here is it:

inline int yyFlexLexer::yylex () { return 0; }; // never called

// end hack

#endif

%}

%option noyywrap nounput noinput batch debug

id    [a-zA-Z][a-zA-Z_0-9]*
int   [0-9]+
blank [ \t\r]

%{
#define YY_USER_ACTION  yylloc->columns (yyleng);
%}

%%

%{
// Code run each time yylex is called.
yylloc->end.filename = &drv.file;
yylloc->step ();
%}

{blank}+	yylloc->step ();
\n+		yylloc->lines (yyleng); yylloc->step ();

"-"		return yy::parser::token::TOK_MINUS;
"+"		return yy::parser::token::TOK_PLUS;
"*"		return yy::parser::token::TOK_STAR;
"/"		return yy::parser::token::TOK_SLASH;
"("		return yy::parser::token::TOK_LPAREN;
")"		return yy::parser::token::TOK_RPAREN;
":="		return yy::parser::token::TOK_ASSIGN;

{int}		{ long n = strtol (yytext, NULL, 10);
		  if (n > INT_MAX || errno == EINVAL || errno == ERANGE)
#ifdef THROW_OK
		    throw yy::parser::syntax_error
		      (*yylloc, "integer is out of range: " +
				std::string(yytext));
#else
		    { std::cerr << *yylloc
			       	<< ": integer is out of range: "
			        << std::string (yytext) << std::endl;
		      exit (EXIT_FAILURE);
		    }
#endif
		  yylval->TOK_NUMBER =  (int) n;
		  return yy::parser::token::TOK_NUMBER;
		  @{ @TOK_NUMBER.Val@ = (int) n; @}
 		}
{id}		{ return yy::parser::token::TOK_IDENTIFIER;
		  @{ @TOK_IDENTIFIER.Name@ = strdup (yytext); @}
		}
.		{
#ifdef THROW_OK
		  throw yy::parser::syntax_error
		    (*yylloc, "invalid character: " + std::string (yytext));
#else
		  std::cerr << *yylloc
 			    << ": invalid character: "
			    << std::string (yytext) << std::endl;
		  exit (EXIT_FAILURE);
#endif
		}
<<EOF>>		{
#ifdef TOK_END
		  return TOK_END; /* #define in `bison' 3.6 .. 3.6.3 GLR skeleton */
#else
		  return yy::parser::token::TOK_END;
#endif
		}
%%
#if defined (USE_FLEXCXX)
#include <fstream>
std::ifstream inputFile;
#else
extern FILE* yyin;
extern int   yy_flex_debug;
#endif

#include <iostream>

void
driver::scan_begin ()
{
  if (file.empty () || file == "-")
    {
      file = "<stdin>";
#if defined (USE_FLEXCXX)
      theScanner = new yoxLexer ();
#else
      yyin = stdin;
#endif
    }
  else
    {
#if defined (USE_FLEXCXX)
      inputFile.open (file.c_str());
      if (inputFile.is_open ())
          theScanner = new yoxLexer (&inputFile, NULL);
      else
#else
      if (!(yyin = fopen (file.c_str (), "r")))
#endif
	{
#if HAVE_STRERROR
	  std::cerr << "cannot open " << file << ": " << strerror(errno) << '\n';
#else
	  perror(("cannot open " + file).c_str ());
#endif
	  exit (EXIT_FAILURE);
	}
    }

#if defined (USE_FLEXCXX)
  theScanner->FlexLexer::set_debug (trace_scanning);
#else
  yy_flex_debug = trace_scanning;
#endif
}

#if defined (USE_FLEXCXX)
// driver::YY_DECL
int driver::yylex (YYLEX_ARGS)
{
  return theScanner->yylex (yylval, yylloc, drv);
}
#endif

void
driver::scan_end ()
{
#if defined (USE_FLEXCXX)
  if (inputFile.is_open ()) inputFile.close ();
#else
  fclose (yyin);
#endif
}
