%option nodefault noinput nounput noyywrap

%option reentrant bison-bridge

%{
#include <errno.h>
#include <limits.h>
#include <stdlib.h>

#include "parser.h"
%}

%%

"="	return TOK_EQ;
"+"	return TOK_PLUS;
"-"	return TOK_MINUS;
"*"	return TOK_STAR;
"/"	return TOK_SLASH;
"^"	return TOK_UP;

"("	return TOK_LPAREN;
")"	return TOK_RPAREN;

[0-9]+	{
	 char leaf_label[100];
	 errno = 0;
	 long n = strtol (yytext, NULL, 10);
	 if (! (INT_MIN <= n && n <= INT_MAX && errno != ERANGE))
	   {yyerror (yyscanner, res, "integer is out of range");
	    if (n < 0) n = INT_MIN; else n = INT_MAX;
	   }
	 yylval->TOK_NUM = (int) n;
	 return TOK_NUM;
	 @{ @TOK_NUM.iVal@ = (int) n;
	    snprintf(leaf_label,sizeof(leaf_label),"NUM|%d",(int) n);
	    agset(@NUM.NIO@, "label", leaf_label);
	 @}
	}

[ \t\f]+ continue;

"\n"	{
	 return TOK_EOL;
	 @{ agset(@TOK_EOL.NIO@, "label", "EOL"); @}
	}

.        yyerror (yyscanner, res, "syntax error, invalid character: %c", yytext[0]);

%%
