%glr-parser %locations
%define api.pure
%define api.prefix {myprefix}
%expect-rr 1

%token CONST TYPE DOTDOT
%token ID
%token NUM

%left '+' '-'

%start decls

%code requires { /* -> header file, before the definition of YYSTYPE */

#include "types.h"

/* MYPREFIXLTYPE is used as an attribute type */
typedef struct MYPREFIXLTYPE {
  int first_line;
  int first_column;
  int last_line;
  int last_column;
} MYPREFIXLTYPE;

/* ignore the default defintion */
#define MYPREFIXLTYPE_IS_DECLARED

}

@autoinh table
@attributes { symTable table; } decl_list
@attributes { symTable table; } decl const_decl const_defs type_decl type_defs
@attributes { symTable table; int iVal; } constant expr NUM
@attributes { char *sym; MYPREFIXLTYPE loc; } ID
@attributes { symTable table; symbol symTabEl; MYPREFIXLTYPE loc; } const_def type_def
@attributes { symTable table; typeSpecT tVal; MYPREFIXLTYPE loc; } type simple_type
@attributes { symTable table; enumListT eVal; MYPREFIXLTYPE loc; } id_list

%code {

#include "util.h"

}

%%

decls	:	decl_list	@{ @i @decl_list.table@ = g_SymbolTable(); @}
      	;

decl_list
	:	decl
	|	decl_list decl
	;

decl	:	const_decl
	|	type_decl
	;

const_decl
	:	CONST const_defs ';'
	;

const_defs
	:	const_def
	|	const_defs ';' const_def
	;

const_def
	:	ID '=' constant	@{ @i @const_def.loc@ = locSpan(@1,@3);
				   @i @const_def.symTabEl@ =
					newConst(@const_def.table@,@const_def.loc@,@ID.sym@,@constant.iVal@);
				@}
	;

constant:	NUM		@{ @i @constant.iVal@ = @NUM.iVal@; @}
	|	'+' NUM		@{ @i @constant.iVal@ = @NUM.iVal@; @}
	|	'-' NUM		@{ @i @constant.iVal@ = -@NUM.iVal@; @}
	|	ID		@{ @i @constant.iVal@ = valOf(@constant.table@,@ID.loc@,@ID.sym@); @}
	|	'+' ID		@{ @i @constant.iVal@ = valOf(@constant.table@,@ID.loc@,@ID.sym@); @}
	|	'-' ID		@{ @i @constant.iVal@ = -valOf(@constant.table@,@ID.loc@,@ID.sym@); @}
	;

type_decl
	:	TYPE type_defs ';'
	;

type_defs
	:	type_def
	|	type_defs ';' type_def
	;

type_def:	ID '=' type	@{ @i @type_def.loc@ = locSpan(@ID.loc@,@type.loc@);
				   @i @type_def.symTabEl@ =
					newType(@type_def.table@,@type_def.loc@,@ID.sym@,@type.tVal@);
				@}
	;

type	:	simple_type	@{ @i @type.loc@ = @simple_type.loc@;
				   @i @type.tVal@ = @simple_type.tVal@;
				@}
	;

	/* the rules for `simple_type' may require the GLR algorithm to disambiguate */
simple_type
	:	'(' id_list ')'	@{ @i @simple_type.loc@ = locSpan(@1,@3);
				   @i @simple_type.tVal@ =
					makeEnumType(@id_list.loc@,@id_list.eVal@);
				@}
	|	expr DOTDOT expr
				@{ @i @simple_type.loc@ = locSpan(@1,@3);
				   @i @simple_type.tVal@ =
					makeSubrangeType(@simple_type.loc@,
							 @expr.0.iVal@,@expr.1.iVal@);
				@}
	|	ID		@{ @i @simple_type.loc@ = @1;
				   @i @simple_type.tVal@ = typeOf(@simple_type.table@,@ID.loc@,@ID.sym@);
				@}
	;

id_list	:	ID		@{ @i @id_list.loc@ = @ID.loc@;
				   @i @id_list.eVal@ = newEnum(@id_list.table@,@ID.loc@,@ID.sym@);
				@}
	|	id_list ',' ID	@{ @i @id_list.0.loc@ = locSpan(@1,@3);
				   @i @id_list.0.eVal@ = addToEnum(@id_list.table@,
								   @ID.loc@,
								   @id_list.1.eVal@,
								   @ID.sym@);
				@}
	;

expr	:	'(' expr ')'	@{ @i @expr.0.iVal@ = @expr.1.iVal@; @}
	|	'+' expr	@{ @i @expr.0.iVal@ = @expr.1.iVal@; @}
	|	'-' expr	@{ @i @expr.0.iVal@ = -@expr.1.iVal@; @}
	|	expr '+' expr	@{ @i @expr.0.iVal@ = @expr.1.iVal@ + @expr.2.iVal@; @}
	|	expr '-' expr	@{ @i @expr.0.iVal@ = @expr.1.iVal@ - @expr.2.iVal@; @}
	|	ID		@{ @i @expr.iVal@ = valOf(@expr.table@,@ID.loc@,@ID.sym@); @}
	|	NUM		@{ @i @expr.iVal@ = @NUM.iVal@; @}
	;

%%
