#include "config.h"

#if HAVE_STRERROR
#include <cstring>
#endif

#include <cstdlib>
#include <cerrno>
#include <cstdio>

#include "parser.hh"
#include "driver.hh"

driver drv;

driver::driver ()
  : trace_parsing (false), trace_scanning (false)
{
  variables["one"] = 1;
  variables["two"] = 2;
}

#if defined (USE_FLEXCXX)
#include <fstream>
std::ifstream inputFile;
#include <FlexLexer.h>
#else
extern FILE* yyin;
extern int   yy_flex_debug;
#endif

#include <iostream>

void
driver::scan_begin ()
{
  if (file.empty () || file == "-")
    {
      file = "<stdin>";
#if defined (USE_FLEXCXX)
      theScanner = new yyFlexLexer ();
#else
      yyin = stdin;
#endif
    }
  else
    {
#if defined (USE_FLEXCXX)
      inputFile.open (file.c_str());
      if (inputFile.is_open ())
	  theScanner = new yyFlexLexer (&inputFile, NULL);
      else
#else
      if (!(yyin = fopen (file.c_str (), "r")))
#endif
	{
#if HAVE_STRERROR
	  std::cerr << "cannot open " << file << ": " << strerror(errno) << '\n';
#else
	  perror(("cannot open " + file).c_str ());
#endif
	  exit (EXIT_FAILURE);
	}
    }

#if defined (USE_FLEXCXX)
  theScanner->set_debug (trace_scanning);
#else
  yy_flex_debug = trace_scanning;
#endif
}

void
driver::scan_end ()
{
#if defined (USE_FLEXCXX)
  if (inputFile.is_open ()) inputFile.close ();
#else 
  fclose(yyin);
#endif
}
