%{
#include <stdio.h>
#include <string.h>
#include "parser.h"
extern	YYLTYPE  yylloc; /* position returned by actions */

void yyerror(YYLTYPE *loc, const char *s);

int yycolumn = 1;

#define set_yylloc(width) { yylloc.first_line = yylineno; \
			    yylloc.first_column = yycolumn; \
			    yycolumn += width; \
			    yylloc.last_line  = yylineno; \
			    yylloc.last_column = yycolumn; }
%}

OctDigit      [0-7]
DecDigit      [0-9]
HexDigit      [0-9a-fA-F]
OctInteger    0({OctDigit})+
DecInteger    ({DecDigit})+
HexInteger    0[xX]({HexDigit})+

%%

[\ \t\f]      { /* ignore whitespace */
		++yycolumn; /* but count columns */ }

[\n]          { yylloc.first_line   = yylineno-1;
		yylloc.first_column = yycolumn-1;
		yycolumn = 1; /* & reset the column */
		yylloc.last_line    = yylineno;
		yylloc.last_column  = yycolumn;

		return('\n'); }

{OctInteger}  { unsigned int oct;
		sscanf(yytext, "%o", &oct);
		yylval.val = (long) oct;
 
		set_yylloc(yyleng);
 
		return(NUMBER);
		@{ @NUMBER.val@ = (long) oct; @} }

{DecInteger}  { int dec;
		sscanf(yytext, "%d", &dec); 
		yylval.val = (long)dec;

		set_yylloc(yyleng);

		return(NUMBER);
		@{ @NUMBER.val@ = (long) dec; @} }

{HexInteger}  { unsigned int hex;
		sscanf(yytext, "%x", &hex);
		yylval.val = (long) hex;

		set_yylloc(yyleng);
 
		return(NUMBER);
		@{ @NUMBER.val@ = (long) hex; @} }

"+"           { set_yylloc(yyleng); return('+'); }
"-"           { set_yylloc(yyleng); return('-'); }
"*"           { set_yylloc(yyleng); return('*'); }
"/"           { set_yylloc(yyleng); return('/'); }
"%"           { set_yylloc(yyleng); return('%'); }
"("           { set_yylloc(yyleng); return('('); }
")"           { set_yylloc(yyleng); return(')'); }

.             { char message[] = "Unrecognized character \"?\"";
		strchr(message, '?')[0] = yytext[0];

		set_yylloc(yyleng);

		yyerror(&yylloc, message); }
%%
int yywrap(void){return 1;}

extern int yydebug;
extern int yyparse(void);

int
main (int argc, const char **argv) {
#if YYDEBUG
  if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

  if (argc > 1)
    yyin = fopen (argv[argc-1], "r");
  else
    yyin = stdin;

  return yyparse();
}
