%option stack
%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "types.h"
#include "lit.h"
#include "name.h"
#include "res.h"
#include "util.h"
#include "parser.h"

/* globals to hold the synthesized attributes for various tokens */
static lit_defn lit;
static name_defn name;

/* HACK: fixed sizes; s.b. malloc'd and then can be realloc'd if needed */
/* accumulate the name of a PLAIN_TAG */
static char plain_tag[1000];
/* accumulate the value of a STRING_DENOTATION */
static char string_denotation[1000];

#define start_plain_tag(s_ptr) {*plain_tag = '\0'; accum_taggle(s_ptr);}
#define accum_taggle(s_ptr)    {accum_tag_name(s_ptr);}

/* HACK: expunge all occurrences of {Intimidate} from the tag */
static void accum_tag_name(char *src) {
  static char intimidate[] = "_";
  char *dest = plain_tag;
  while (*src != '\0') {
    int i = strcspn(src, intimidate);
    if (i > 0) {strncat(dest, src, i); src += i;}
    if (*src++ == '\0') break; /* skip {Intimidate} */
  }
}

#define start_string_denotation(pch) {*string_denotation='\0'; accum_string_denotation(pch);}
#define accum_string_denotation(pch) {accum_C_literal(pch);}

static void
accum_C_literal(char *pch) {
  strcat(string_denotation, pch);
}
%}

	/* regular expression definitions */

Lower             [a-z]
Upper             [A-Z]
Letter            {Lower}|{Upper}
Digit             [0-9]
ExtendedDigit     {Digit}|[a-fA-F]
LowerDigit        {Lower}|{Digit}
UpperDigit        {Upper}|{Digit}
LetterDigit       {Letter}|{Digit}

FirstTaggle       {Lower}{Intimidate}?({LowerDigit}+{Intimidate}?)*
Taggle            ({LowerDigit}+{Intimidate}?)+

WhiteSpace        [\ \t\f\n\r]
TagConj           [\ \t\f\n\r]
TagDisj           [^a-z0-9\_\t\f\n\r]

Intimidate        "_"

BoldTag           {Upper}{UpperDigit}*
Disj              [^A-Z0-9\_]

PragmatTag        PR(AGMAT)?

CommentTag        CO(MMENT)?

IntegerLiteral    {Integer}
RealLiteral       {RealLiteral1}|{RealLiteral2}
RealLiteral1      {Integer}{Exponent}
RealLiteral2      {Integer}?{Fraction}{Exponent}?
Integer           {Digit}+
Fraction          "."{Integer}
/* \ is allowed in the Revised Report as an alternative */
Exponent          [eE\\][-+]?{Integer}

RadixLiteral      {Radix}{RadixInteger}
Radix             (2|4|8|16)[rR]
RadixInteger      {ExtendedDigit}+

ApostropheImage   \'\'
QuoteImage        \"\"
BackSpaceImage    \'[bB][sS]
NewPageImage      \'[fF][fF]
NewLineImage      \'[nN][lL]
ReturnImage       \'[cC][rR]
HorizTabImage     \'[hH][tT]
VertTabImage      \'[vV][tT]

%s UPPER
%x PRAG
%x CO
%x TAG

%x BRIEF_CO
%x STRING STRBRK

%%

[\ \t\f\r]      {/* ignore space, tab, formfeed and return */}
\n              {/* ignore newline */}

%{
  /* categories of lexical state machines for program text:
     UPPER: top level program text within particular_program;
                      brief_enclosed_clause */
%}
"("		return ('(');
")"		return (')');
"["		return ('[');
"]"		return (']');
":"		return (':');
";"		return (';');
","		return (',');
"="		return ('=');
"~"		return ('~');
":="		return (_BECOMES);
%{
  /* rules for operators */
%}
"^"		return ('^'); /* Monadic */
"&"		return ('&'); /* Monadic */
"%"		return ('%'); /* Monadic */
"+"		return ('+'); /* Monadic & Dyadic */
"-"		return ('-'); /* Monadic & Dyadic */
">"		return ('>'); /* Dyadic */
"<"		return ('<'); /* Dyadic */
"/"		return ('/'); /* Dyadic */
"*"		return ('*'); /* Dyadic */
%{
  /* rules for pragmats and comments */
%}
{PragmatTag}/{Disj}     {/* start a PRAGMAT */
                         /* push start condition, switch to PRAG */
                         yy_push_state ( PRAG );
                         yymore(); /* accumulate contents of PRAGMAT */ }
<PRAG>{PragmatTag}/{Disj} {/* end the PRAGMAT; the contents are
                            in yytext, with surrounding {PragmatTag}s */
                         /* pop start condition stack, switch to it */
                         yy_pop_state (); }
<PRAG>\n|.              {yymore(); /* accumulate contents of PRAGMAT */ }
\#                      {/* start a # */
                         /* push start condition, switch to BRIEF_CO */
                         yy_push_state ( BRIEF_CO ); }
{CommentTag}/{Disj}     {/* start a COMMENT */
                         /* push start condition, switch to CO */
                         yy_push_state ( CO ); }
<BRIEF_CO>#             {/* end the # */
                         /* pop start condition stack, switch to it */
                         yy_pop_state (); }
<CO>{CommentTag}/{Disj} {/* end the COMMENT */
                         /* pop start condition stack, switch to it */
                         yy_pop_state (); }
<CO,BRIEF_CO>\n|.       ;
%{
  /* rules for tags */
%}
{BoldTag}/{Disj}        {/* we've got a BoldTag; check if it's a special BoldTag */
                         const res_word *res;
                         if ( ( res = in_res_word_set ( yytext, yyleng ) )
                           && ( res->kind == keyword || res->kind == primitive_mode ) )
                           return ( res->token );
                         /* no, ensure it's in the indicator table */
                         name = add_name ( yytext );
                         return (_BOLD_TAG);
                         @{ @_BOLD_TAG.name@ = name; @} }
%{
  /* Following PLAIN_TAG processing isn't quite correct:
     - rules for {Intimidate} may not be quite right */
%}
{FirstTaggle}/{TagConj} {/* the start of a Tag */
			 start_plain_tag ( yytext );
			 yy_push_state ( TAG ); /* TAG processing */ }
{FirstTaggle}/{TagDisj} {/* a Tag */
			 start_plain_tag ( yytext );
			 name = add_name ( plain_tag );
			 s_ident_name ( name );
			 return (_PLAIN_TAG);
			 @{ @_PLAIN_TAG.name@ = name; @} }
<TAG>{Taggle}/{TagConj} {/* the next Taggle */
                         accum_taggle ( yytext ); }
<TAG>{Taggle}/{TagDisj} {/* the last Taggle */
                         accum_taggle ( yytext );
                         /* terminate the Tag being accumulated */
                         name = add_name ( plain_tag );
                         s_ident_name ( name );
                         /* pop start condition stack, switch to it */
                         yy_pop_state (); /* prior start condition */
                         return (_PLAIN_TAG);
                         @{ @_PLAIN_TAG.name@ = name; @} }
<TAG>\n|{TagConj}       ; /* Tag may include [\n\ \t\f\r] characters;
			     stripped out of canonical representation */
<TAG>{TagDisj}          {/* already had the last Taggle; push back {TagDisj} */
#if 0
                         /* Copy yytext because unput() trashes yytext */
                         char c = *yytext; unput( c ); /* rescan the {Disj} */
#else
			 yyless(0); /* rescan the {TagDisj} */
#endif
                         /* terminate the Tag being accumulated */
                         name = add_name ( plain_tag );
                         s_ident_name ( name );
                         /* pop start condition stack, switch to it */
                         yy_pop_state (); /* prior start condition */
                         return (_PLAIN_TAG);
                         @{ @_PLAIN_TAG.name@ = name; @} }
<TAG>.                  {/* anything else is unexpected */
                         fprintf(stderr, "unexpected \"%s\" @ [%d].\n", yytext, yylineno);}
%{
  /* rules for denotations */
%}
{IntegerLiteral}        {lit = add_lit(yytext);
                         s_int_lit(lit, yytext);
                         return (_INTEGRAL_DENOTATION);
                         @{ @_INTEGRAL_DENOTATION.lit@ = lit; @} }
{RealLiteral}           {lit = add_lit(yytext);
                         s_real_lit(lit, yytext);
                         return (_REAL_DENOTATION);
                         @{ @_REAL_DENOTATION.lit@ = lit; @} }
{RadixLiteral}          {lit = add_lit(yytext);
                         s_bits_lit(lit, yytext);
                         return (_BITS_DENOTATION);
                         @{ @_BITS_DENOTATION.lit@ = lit; @} }
\"                      {/* Start of a {StringLiteral} */
                         start_string_denotation(yytext);
                         yy_push_state(STRING); /* STRING processing */ }
<STRING>[^"'\b\f\n\r\t\v\\] accum_string_denotation(yytext);
<STRING>[\b]|{BackSpaceImage} accum_string_denotation("\\b");
<STRING>[\f]|{NewPageImage} accum_string_denotation("\\f");
<STRING>[\n]            accum_string_denotation("\\n");
<STRING>{NewLineImage}  accum_string_denotation("\\n");
<STRING>[\r]|{ReturnImage} accum_string_denotation("\\r");
<STRING>[\t]|{HorizTabImage} accum_string_denotation("\\t");
<STRING>[\v]|{VertTabImage} accum_string_denotation("\\v");
<STRING>[\\]            accum_string_denotation("\\\\");
<STRING>{ApostropheImage} accum_string_denotation("\\\'");
<STRING>{QuoteImage}    accum_string_denotation("\\\"");
<STRING>\"/{WhiteSpace}+\" {/* Start of a {StringBreak} */
                         yy_push_state(STRBRK); /* STRBRK processing */}
<STRING>\"              {/* End of a {StringLiteral} */
                         accum_string_denotation(yytext);
                         /* terminate the {StringLiteral} being accumulated */
                         lit = add_lit(string_denotation);
                         s_string_lit(lit, string_denotation);
                         /* pop start condition stack, switch to it */
                         yy_pop_state(); /* prior start condition */
                         return (_STRING_DENOTATION);
                         @{ @_STRING_DENOTATION.lit@ = lit; @}}
<STRING>.               fprintf(stderr, "unexpected \"%s\" @ [%d].\n", yytext, yylineno);
<STRBRK>\n              ;
<STRBRK>{WhiteSpace}    ; /* still in a {StringBreak} */
<STRBRK>\"              {/* End of a {StringBreak} */
                         /* pop start condition stack, switch to it */
                         yy_pop_state(); /* prior start condition */}
%{
  /* anything else is unexpected */
%}
.                       fprintf(stderr, "unexpected \"%s\" @ [%d].\n", yytext, yylineno);

%%
int yywrap(void){return 1;}
