/*
 * util.c - commonly used utility functions.
 */

#if defined (__cplusplus)
#include <cstdio>
#include <cstdlib>
#include <cstring>
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "types.h"
#include "util.h"

void *checked_malloc(size_t len) {
  void *p = malloc(len);
  if (!p) {
     fprintf(stderr, "\nRan out of memory!\n");
     abort();
  }
  return p;
}

void *checked_calloc(size_t count, size_t size) {
  void *p = calloc(count, size);
  if (!p) {
     fprintf(stderr, "\nRan out of memory!\n");
     abort();
  }
  return p;
}

string toString ( char *s ) { 
  string p = ( string ) checked_malloc ( strlen(s)+1 );
  strcpy(p,s);
  return p;
}

/* make a mode_defn value */
mode_defn mkPrimitiveMode(enum mode which) {
  simple_mode_defn simple = {which};
  mode_defn new_mode_defn; new_mode_defn._simple = simple;
  return new_mode_defn;
}

mode_defn mkSithetyMode(enum mode which, int num_short, int num_long) {
  sithety_mode_defn sithety = {which, {num_short, num_long}};
  mode_defn new_mode_defn; new_mode_defn._sithety = sithety;
  return new_mode_defn;
}

mode_defn mkUnitedMode(mode_defn *mode_set) {
  mode_defn new_mode_defn;
  new_mode_defn._united.tag = union_mode;
  new_mode_defn._united.decl_list = mode_set;
  return new_mode_defn;
}

mode_defn mkStructureMode(field **field_set) {
  mode_defn new_mode_defn;
  new_mode_defn._struct.tag = struct_mode;
  new_mode_defn._struct.field_list = field_set;
  return new_mode_defn;
}

mode_defn mkArrayMode(void *rowers, mode_defn elem_mode) {
  mode_defn new_mode_defn;
  new_mode_defn._array.tag        = array_mode;
  new_mode_defn._array.rower_list = NULL;
  return new_mode_defn;
}

mode_defn mkProcMode(mode_defn *params, mode_defn ret_mode) {
  mode_defn new_mode_defn;
  new_mode_defn._proc.tag         = proc_mode;
  new_mode_defn._proc.params      = params;
  return new_mode_defn;
}

mode_defn mkRefMode(mode_defn deref_mode) {
  mode_defn new_mode_defn;
  new_mode_defn._ref.tag          = ref_mode;
  return new_mode_defn;
}

mode_defn balanceMode(mode_defn left, mode_defn right) {
  return left; /* no balancing for now */
}

/* lookup a mode_defn value in the global symbol table */
mode_defn gMode(name_defn name) {
  return g_mode(name);
}
