/* $Id: joinTabs.c 2098 2021-02-16 02:47:30Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

#if defined (__cplusplus)
#include <cstdio>
#include <cstdlib>
#include <cstring>
#else
#include <stdio.h> 
#include <stdlib.h> 
#include <string.h>
#endif

#include "glob.h" 
#include "opts.h" 


extern long *indToEntsByProdNum; 
extern int prodNumSeq; 
extern FILE *yoxOutG; 
extern int permitCtimeCircTest; 
extern long termSeq; 
extern long *leafRefCountInitIndex,*leafRefCountListLength; 
extern long eeCountInitSeq; 
extern long attrbOccrSeq; 
extern long entOccrSeq; 
extern long leafRefCountInitSeq; 

extern void compilerTextCopy1(void); 
extern void compilerTextCopy2(void); 
extern void compilerTextCopy3(void); 

FILE *temp1Out; 
FILE *temp2Out; 
FILE *temp3Out; 
FILE *temp4Out; 
FILE *temp5Out; 
FILE *temp6Out; 
FILE *temp7Out; 
FILE *temp8Out; 
FILE *tempTrashOut;
FILE *tempTravOut;

#define xferBufSize 1024 
char xferBuf[xferBufSize];




void openTempFiles(void)
  {temp1Out=sfopen(temp1FileID,"w"); 
   temp2Out=sfopen(temp2FileID,"w"); 
   temp3Out=sfopen(temp3FileID,"w"); 
   temp4Out=sfopen(temp4FileID,"w"); 
   temp5Out=sfopen(temp5FileID,"w"); 
   temp6Out=sfopen(temp6FileID,"w"); 
   temp7Out=sfopen(temp7FileID,"w"); 
   temp8Out=sfopen(temp8FileID,"w"); 
   tempTrashOut=sfopen(tempTrashFileID,"w"); 
   tempTravOut=sfopen(tempTravFileID,"w"); 
  }



void erasf(const char *str) 
  {char dummy[64];  
   strcpy(dummy,"rm "); 
   if (system(strcat(dummy,str))) 
      fprintf(stderr,"ox: system error removing %s",str); 
  } 

 

void rmTempFiles(void) 
  {fclose(temp1Out); 
   fclose(temp2Out); 
   fclose(temp3Out); 
   fclose(temp4Out); 
   fclose(temp5Out); 
   fclose(temp6Out); 
   fclose(temp7Out); 
   fclose(temp8Out); 
   fclose(tempTrashOut); 
   fclose(tempTravOut); 
   erasf(temp1FileID);    
   erasf(temp2FileID);    
   erasf(temp3FileID);    
   erasf(temp4FileID);    
   erasf(temp5FileID);    
   erasf(temp6FileID);    
   erasf(temp7FileID);    
   erasf(temp8FileID);    
   erasf(tempTrashFileID);    
   erasf(tempTravFileID);    
  }



/* Closes an open file, reopens it for reading, copies it to a file 
 * that is already open for writing. 
 */ 
void appendToOutFile(FILE *tabFile, const char *tabFileStr, FILE *dest) 
  {long count; 
   fclose(tabFile); 
   tabFile = fopen(tabFileStr,"r"); 
   while ((count = fread(xferBuf,sizeof(char),xferBufSize,tabFile)))    
      fwrite(xferBuf,sizeof(char),count,dest); 
  } 



void organizeTables(void) 
  {int i; 
   fprintf(yoxOutG,"\n\n#include <limits.h> "); 
   compilerTextCopy1(); 
   fprintf(yoxOutG,"\n\n\n\nyyyFT yyyRCIL["); 
   if (eeCountInitSeq) 
      {fprintf(yoxOutG,"] = {\n"); 
       appendToOutFile(temp2Out,temp2FileID,yoxOutG);  
       fprintf(yoxOutG,"\n}"); 
      }  
      else 
      fprintf(yoxOutG,"1]"); 
   fprintf(yoxOutG,";\n"); 

   fprintf(yoxOutG,"\nshort yyyIIIEL[] = {0,\n"); 
   for (i=1;i<prodNumSeq;i++) 
     {fprintf(yoxOutG,"%ld,",indToEntsByProdNum[i]);   
      if (!(i%10)) fprintf(yoxOutG,"\n");
     }  
   fprintf(yoxOutG,"\n};\n"); 

   fprintf(yoxOutG,"\nlong yyyIIEL[] = {\n"); 
   appendToOutFile(temp5Out,temp5FileID,yoxOutG);  
   fprintf(yoxOutG,"\n};\n"); 

   fprintf(yoxOutG,"\nlong yyyIEL["); 
   if (attrbOccrSeq) 
      {fprintf(yoxOutG,"] = {\n"); 
       appendToOutFile(temp4Out,temp4FileID,yoxOutG);  
       fprintf(yoxOutG,"%ld,",entOccrSeq*2); 
       fprintf(yoxOutG,"\n}"); 
      }
      else 
      fprintf(yoxOutG,"1]"); 
   fprintf(yoxOutG,";\n"); 

   fprintf(yoxOutG,"\nyyyFT yyyEntL["); 
   if (entOccrSeq) 
      {fprintf(yoxOutG,"] = {\n"); 
       appendToOutFile(temp3Out,temp3FileID,yoxOutG);  
       fprintf(yoxOutG,"\n}"); 
      }
      else  
      fprintf(yoxOutG,"1]"); 
   fprintf(yoxOutG,";\n"); 

   compilerTextCopy2(); 

   fprintf(yoxOutG,"\n\n\n"); 
   appendToOutFile(temp1Out,temp1FileID,yoxOutG);  
   
   if (permitCtimeCircTest) 
      {
       fprintf(yoxOutG,"\n\n\nyyyWAT yyyLRCIL[%ld]",
                       (leafRefCountInitSeq?leafRefCountInitSeq:1)  
              ); 
       if (leafRefCountInitSeq) 
          {fprintf(yoxOutG," = {"); 
           appendToOutFile(temp7Out,temp7FileID,yoxOutG);  
           fprintf(yoxOutG,"\n}"); 
          } 
       fprintf(yoxOutG,";\n"); 
      }  
   compilerTextCopy3(); 
  } 
  
 
