/* $Id: mac.c 2098 2021-02-16 02:47:30Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

/* abbreviations in variable names: 
 * mac        macro  
 * tab        table
 * sp         space
 * sz         size
 * frm        formal
 * act        actual
 * prm        parameter
 * str        string
 * stg        storage
 * nx         next
 * fr         free
 * MD         macro define
 * MU         macro use
 * LLA        lexer lookahead 
 */ 

#if defined (__cplusplus)
#include <cstdlib>
#include <cstdio>
#include <cstring>
#else
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#endif

#include "opts.h"
#include "scan.h"


#define frmPrmIndFormat "#%4d#" 
#define fpifLen 6  
#define currBodyStart ((macTab[macStackTop->defNum]).bodyStart)
#define currAfterBody ((macTab[macStackTop->defNum]).afterBody)
#define inFromActPrm ((macStackTop->nextChar >= macActPrmStg) \
                      && \
                      (macStackTop->nextChar < afterMacActPrmStg) \
                     ) 



struct macroDef {char *name;
                 char **firstFrmPrm;
                 char **afterFrmPrms;
                 char *bodyStart;
                 char *afterBody;
                };


struct macroUse {int defNum;
                 char **firstActPrm;
                 char *nextChar;
                 char *bodyPtr;
                };


int inFromMac = 0; 
char tempLLAbuf[BUFSIZ];
char *LLAbufStart,*LLAbufNxFr; 



void switchInputStreamToMacro(char *cp1, char *cp2)
  {char *cpDumFrom,*cpDumTo; 

   inFromMac = 1; 
   cpDumFrom = LLAbufStart = cp1; 
               LLAbufNxFr  = cp2; 
   cpDumTo = tempLLAbuf; 
   while (cpDumFrom < LLAbufNxFr) *cpDumTo++ = *cpDumFrom++;  
  }



void switchInputStreamFromMacro(void)
  {char *cpDumFrom,*cpDumTo; 

   inFromMac = 0; 
   cpDumTo = LLAbufStart;
   cpDumFrom = tempLLAbuf; 
   while (cpDumTo < LLAbufNxFr) *cpDumTo++ = *cpDumFrom++;  
  }



int actPrmCount; 


long macTabSz = 32;
long macFrmPrmSpSz = 128; 
long macBodyStgSz = 16384;

struct macroDef *macTab,*afterMacTab,*macTabNxFr;
char **macFrmPrmSp,**afterMacFrmPrmSp,**macFrmPrmSpNxFr; 
char *macBodyStg,*afterMacBodyStg,*macBodyStgNxFr; 

void prepareDataStructuresMD(void)
  {
   if ((macTab = (struct macroDef *)malloc(macTabSz * sizeof(struct macroDef)))
       == 
       (struct macroDef *)NULL
      )
      fatal("malloc error in macro definition table allocation\n");  
   afterMacTab = (macTabNxFr = macTab) + macTabSz; 
   
   if ((macFrmPrmSp = (char **)malloc(macFrmPrmSpSz * sizeof(char *)))
       == 
       (char **)NULL
      )
      fatal("malloc error in macro formal parameter space allocation\n");  
   afterMacFrmPrmSp = (macFrmPrmSpNxFr = macFrmPrmSp) + macFrmPrmSpSz; 
   
   if ((macBodyStg = (char *)malloc(macBodyStgSz))
       == 
       (char *)NULL
      )
      fatal("malloc error in macro body storage allocation\n");  
   afterMacBodyStg = (macBodyStgNxFr = macBodyStg) + macBodyStgSz; 
  }



long macStackSz = 16;
long macActPrmSpSz = 128; 
long macActPrmStgSz = 2048; 

struct macroUse *macStack,*afterMacStack,*macStackTop; 
char **macActPrmSp,**afterMacActPrmSp,**macActPrmSpNxFr; 
char *macActPrmStg,*afterMacActPrmStg,*macActPrmStgNxFr; 

void prepareDataStructuresMU(void)
  {
   if ((macStackTop = 
          (struct macroUse *)malloc(macStackSz * sizeof(struct macroUse))
       )
       == 
       (struct macroUse *)NULL
      )
      fatal("malloc error in macro stack allocation\n");  
   macStack = macStackTop + 1; 
   afterMacStack = macStackTop + macStackSz; 
   
   if ((macActPrmSp = (char **)malloc(macActPrmSpSz * sizeof(char *)))
       == 
       (char **)NULL
      )
      fatal("malloc error in macro actual parameter space allocation\n");  
   afterMacActPrmSp = (macActPrmSpNxFr = macActPrmSp) + macActPrmSpSz; 
   
   if ((macActPrmStg = (char *)malloc(macActPrmStgSz))
       == 
       (char *)NULL
      )
      fatal("malloc error in macro actual parameter storage allocation\n");
   afterMacActPrmStg = (macActPrmStgNxFr = macActPrmStg) + macActPrmStgSz; 
  }



int macroLookup(char *cp)
  {struct macroDef *mdDum; 
   for (mdDum = macTab; mdDum<macTabNxFr; mdDum++)
     if (mdDum->name == cp) 
         return(mdDum - macTab); 
   return -1;
  }



void initMacroDef(char *cp)
  {if (macroLookup(cp) != -1) 
      {yerror("macro multiply defined:  "); 
       fprintf(stderr,"%s\n",cp); 
      }
   if (macTabNxFr == afterMacTab) handleOverflow(NMCD);
   macTabNxFr->name = cp;
   macTabNxFr->firstFrmPrm = macTabNxFr->afterFrmPrms = macFrmPrmSpNxFr;
   macTabNxFr->bodyStart = macTabNxFr->afterBody = macBodyStgNxFr; 
  } 



int macroFrmPrmLookup(char *cp) 
  {char** fpDum;
   for (fpDum = macTabNxFr->firstFrmPrm; 
        fpDum < macTabNxFr->afterFrmPrms; 
        fpDum++ 
       )
     if (*fpDum == cp) 
        return(fpDum - macTabNxFr->firstFrmPrm); 
   return -1;
  } 



void installMacroFrmPrm(char *cp)
  {if (macroFrmPrmLookup(cp) != -1) 
      ywarning("duplicate formal parameter in macro\n");  
   if (macFrmPrmSpNxFr == afterMacFrmPrmSp) handleOverflow(NMFP); 
   *macFrmPrmSpNxFr++ = cp;
   macTabNxFr->afterFrmPrms = macFrmPrmSpNxFr;
  } 



void putIDinMacroDefBody(char *cp) 
  {int index,len; 
   if ((index = macroFrmPrmLookup(cp)) == -1) 
      {if ((macBodyStgNxFr + (len = strlen(cp))) >= afterMacBodyStg)  
          handleOverflow(MBSS); 
          else 
          {strcpy(macBodyStgNxFr,cp); 
           macBodyStgNxFr += len; 
          }
      }
      else 
      {if ((macBodyStgNxFr + fpifLen) >= afterMacBodyStg)  
          handleOverflow(MBSS); 
          else 
          {sprintf(macBodyStgNxFr,frmPrmIndFormat,index); 
           macBodyStgNxFr += fpifLen;   
          }
      }
   macTabNxFr->afterBody = macBodyStgNxFr; 
  }



void putCharInMacroDefBody(char c) 
  {int i; 
   
   for (i=0; i <= (c == '#'); i++)
     {if (macBodyStgNxFr == afterMacBodyStg) handleOverflow(MBSS); 
      *macBodyStgNxFr++ = c; 
     } 
   macTabNxFr->afterBody = macBodyStgNxFr; 
  }



void terminateMacroDef(void)
  {macTabNxFr++; 
  } 



void macroInterpSetup1(int i) 
  {actPrmCount = 0; 
   if (++macStackTop == afterMacStack) handleOverflow(MSSZ);  
   macStackTop->defNum = i; 
   macStackTop->firstActPrm = macActPrmSpNxFr;
   if (macActPrmSpNxFr == afterMacActPrmSp) handleOverflow(NMAP);   
   *macActPrmSpNxFr = macActPrmStgNxFr; 
  }



int macroInterpSetup1IfMacro(char *cp)
  {int i; 

   if ((i = macroLookup(cp)) == -1) 
      return 0; 
      else 
      {macroInterpSetup1(i);
       return 1;
      }
  }



void putCharInMacroActParam(char c) 
  {if (macActPrmStgNxFr == afterMacActPrmStg) handleOverflow(MAPS); 
   *macActPrmStgNxFr++ = c; 
  }



void terminateActualParam(void)
  {putCharInMacroActParam('\0'); 
   if (++macActPrmSpNxFr == afterMacActPrmSp) handleOverflow(NMAP);   
   *macActPrmSpNxFr = macActPrmStgNxFr; 
   actPrmCount++; 
  } 



void returnFromMacro(void)
  {macActPrmSpNxFr = macStackTop->firstActPrm;
   macActPrmStgNxFr = *(macStackTop->firstActPrm);
   macStackTop--;
  }



/* entered with macStackTop->nextChar pointing to the character after 
 * the one last gotten from the macro input stream, if one was just 
 * gotten, else points to the first character in the body of a 
 * macro use just placed on the macro stack.  Returns true if exiting
 * with macStackTop->nextChar pointing to a valid character, else 
 * returns false and the macro use stack is empty.  
 */ 
int makeReadyInputFromMacro(void)
  {int index; 

   do 
     {if (inFromActPrm)
         {if (*(macStackTop->nextChar) == '\0') 
             {macStackTop->nextChar = macStackTop->bodyPtr;
              continue; 
             } 
         } 
         else
         {if (macStackTop->nextChar == currAfterBody) 
             {returnFromMacro(); 
              continue; 
             }
             else
             if (*(macStackTop->nextChar) == '#') 
                {if (*(macStackTop->nextChar + 1) == '#') 
                    (macStackTop->nextChar)++;
                    else 
                    {sscanf(macStackTop->nextChar,frmPrmIndFormat,&index);
                     macStackTop->bodyPtr = (macStackTop->nextChar) + fpifLen;
                     macStackTop->nextChar = 
                       *(macStackTop->firstActPrm + index);
                     continue;  
                    }
                }
         } 
      return 1; 
     } 
     while (macStackTop >= macStack); 
   return 0; 
  }



void macroInterpSetup2(char *cp1, char *cp2)
  {int fpCount; 
   struct macroDef mdDum; 
   mdDum = macTab[macStackTop->defNum];
   fpCount = (mdDum.afterFrmPrms - mdDum.firstFrmPrm);
   if (actPrmCount > fpCount)  
      yerror("too many actual parameters in macro invocation.\n"); 
   if (actPrmCount < fpCount)  
      fatal("too few actual parameters or missing comma in macro invocation.\n"
           ); 
   macStackTop->nextChar = mdDum.bodyStart; 
   if (makeReadyInputFromMacro()) switchInputStreamToMacro(cp1,cp2);  
  }



/* This is called by input() exactly when inFromMac is set.  
 * macStackTop->nextChar always points to a valid character (in an 
 * actual parameter or in the body of the most recently invoked macro) when 
 * this is called.  So a valid character is always returned by this
 * routine.  Before returning, macStackTop->nextChar must be set to  
 * point to a valid character or the macro invocation stack must be 
 * emptied and inFromMac reset, which functions are performed by the 
 * call to makeReadyInputFromMacro().  
 */  
char inputFromMacro(void)
  {char c;

   c = *(macStackTop->nextChar)++; 
   if (!makeReadyInputFromMacro()) switchInputStreamFromMacro();  
   if(c=='\n')c='\t';
   return c;
  }



