/* $Id: scan.c 2625 2022-08-02 02:07:14Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

#if defined (__cplusplus)
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cctype>
#else
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#endif

#include "glob.h" 
#include "scan.h" 
#include "YYLENG_INT.h" 
extern void flushAll(void);

int whichScanner; 
int *plineno; 
#if YYLENG_INT
int *pleng; 
#else
size_t *pleng;
#endif
char *ptext; 
char *inFileStr; 
int removeOutput = 0; 

int line; 
int nLbraces; 
int nLparens; 

char *IDbuffer; 

char *gsBuffer; 
int suffixBuffer;  
char *atBuffer; 



int firstMark = 1;  
int alreadyAugmented = 0; 
char *startSymLexeme = (char *)NULL; 
char *dumNodeSym; 


#if defined (USING_REFLEX)
int yywrap(void) 
  {switch(whichScanner) 
     {case 1: return yy1wrap(); 
      case 2: return yy2wrap(); 
      case 3: return yy3wrap(); 
      case 4: return yy4wrap(); 
      default: fatal("request for unknown scanner (yywrap)\n");
     } 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 
#endif



int yylex(void) 
  {switch(whichScanner) 
     {case 1: return yy1lex(); 
      case 2: return yy2lex(); 
      case 3: return yy3lex(); 
      case 4: return yy4lex(); 
      default: fatal("request for unknown scanner\n");
     } 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 


/*
int yylex(void) 
  {
   int temp;  
   fprintf(stderr,"\ncalling yy%clex\n",'0' + whichScanner); 
   switch(whichScanner) 
     {case 1: temp = yy1lex();break;
      case 2: temp = yy2lex();break;
      case 3: temp = yy3lex();break;
      default: fatal("request for unknown scanner\n");
     } 
              fprintf(stderr,"%d returned from yy%clex, line %d\n",
                      temp,'0' + whichScanner,*plineno); 
              return temp; 
  } 
*/



void printAugmentationProduction(void) 
  {if (startSymLexeme == (char *)NULL) startSymLexeme = IDbuffer; 
   fprintf(yoxOutG,"%s \n\t:\t{if (!yyyInitDone) \n", 
                    augNstr
          ); 
   fprintf(yoxOutG,"\t\t    {yyyYoxInit(); \n");  
   fprintf(yoxOutG,"\t\t     yyyInitDone = 1;\n"); 
   fprintf(yoxOutG,"\t\t    }\n"); 
   fprintf(yoxOutG,"\t\t yyyYoxReset();\n\t\t}\n\t\t%s\n",
                   startSymLexeme 
          ); 
   fprintf(yoxOutG,
      "\t\t{yyyDecorate(); yyyExecuteRRsection(); } \n\t;\n\n"
          ); 
   alreadyAugmented = 1; 
  } 
 


void scanAttrbName(char *ptext, int leng) 
  {int charConstSym;  
   int i; 

   suffixBuffer = 0; 
   charConstSym = (ptext[0] == '\'')?1:0;  
   i = leng - 1; 
   while ((i >= 0) && (ptext[i] != '.')) i--; 
   if (((i == -1) || (i == (leng - 1))) ||  
       ((atBuffer = stringLookup(ptext+i+1,leng - i - 1,0)) == (char *)NULL) 
      ) 
      goto err; 
      else 
      ptext[i--] = ' ';  
   if (!charConstSym)   
      {if ((gsBuffer = stringLookup(ptext,i+1,0)) != (char *)NULL) return; 
       if (ptext[i] == '.') goto err;  
       while ((i >= 0) && (ptext[i] != '.')) 
             if (!isdigit(ptext[i])) goto err; else  i--;  
       if (i == -1)  goto err;   
       if ((gsBuffer = stringLookup(ptext,i,0)) != (char *)NULL) 
          {sscanf(ptext+i+1,"%d",&suffixBuffer);  
           return; 
          } 
          else goto err;  
      } 
      else  
      {if (ptext[i] == '\'')  
          {gsBuffer = getFirstTextOfCC(stringLookup(ptext,i+1,1));  
           return; 
          }  
       if (!isdigit(ptext[i])) goto err;  
       while ((i >= 0) && (isdigit(ptext[i]))) i--;  
       if ((i == -1) || (ptext[i] != '.')) goto err; else ptext[i--] = ' ';   
       if (ptext[i] == '\'')  
          {gsBuffer = getFirstTextOfCC(stringLookup(ptext,i+1,1));   
           sscanf(ptext+i+1,"%d",&suffixBuffer);  
           return; 
          } 
      } 
err:
   {yerror("syntax error or unknown identifier in attribute name:\n"); 
    fprintf(stderr,"     %s \n",ptext); 
    suffixBuffer = -1; 
    return;  
   } 
  } 


/* Print a string along with the line and column number of the first 
   character of the prefix of the remaining input that has just caused 
   an error.  Called by yyparse().  See section 8 ("The Yacc Environment")   
   of the yacc document.  
*/
void yerror (const char *msg) 
  {fprintf(stderr,"ox: %s: line %d: %s",inFileStr,*plineno,msg);  
   flushAll();
   removeOutput = 1;
  }
 
/* void yerror(int nMsgs, ...)  
  {int i;
   char *msg;
   va_list ap; 

   va_start(ap,nMsgs);   
   fprintf(stderr,"ox: %s: line %d:  ",inFileStr,*plineno);  
   for (i=0;i<nMsgs;i++) 
       {msg = va_arg(ap,char *); 
        fprintf(stderr,"%.80s",msg);
       }  
   flushAll(); 
   va_end(ap); 
  } 
*/ 

void ywarning (const char *msg) 
  { 
   fprintf(stderr,"ox: %s: warning: line %d: %s",inFileStr,*plineno,msg);  
   flushAll(); 
  }
 

/* Print a string and exit */ 
void fatal(const char *msg)    
  {yerror(msg); 
   exit(-1);    
  } 



#if YYBTYACC && !YYBYACC
void yyerror(const char *msg, ...)
#else  
void yyerror(const char *msg)
#endif
  {if (strcmp(msg,"syntax error")) yerror(msg);
       else
       yerror("syntax error\n"); 
  } 



int getScnState(void) 
  {switch(whichScanner) 
     {case 1: return yy1ScnState(); 
      case 2: return yy2ScnState(); 
      case 3: return yy3ScnState(); 
      default: fatal("request for unknown scanner\n");
     } 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



void debCom(void) 
  {fprintf(stderr,"eating comment at line %5d in scanner %d in sc %5d\n", 
	   *plineno,whichScanner,getScnState());   
  } 


 
void debString(void) 
  {fprintf(stderr,"eating string at line %5d in scanner %d in sc %5d\n", 
	   *plineno,whichScanner,getScnState());   
  } 


 
void debReportSC(const char *str) 
  {fprintf(yoxOutG,"\n%s: in scanner %5d in sc %5d    ", 
	           str,whichScanner,getScnState()
          );  
  } 



/* EOGFRET becomes the start condition upon the first execution of yywrap(). 
 * Its only purpose is to cause the scanner to return a token to signify 
 * to the parser that the end of the file containing the syntactic 
 * specification has been encountered.  
*/ 
 

