/* To demonstrate the use of an attribute grammar for the printing of parse
 *  trees of a simple expression grammar.
 */

%{
#if defined (__cplusplus)
#include <cstdio>
#include <cstring>
#else /* !defined (__cplusplus) */
#include <stdio.h>
#if defined (__sun) && defined (__SVR4)
#include <strings.h>
#else
#include <string.h>
#endif
#endif /* !defined (__cplusplus) */

#include "pt.h"

#if YYBTYACC && !YYBYACC
/* the `btyacc' 3.0 skeleton omits the following */
#include "parser.h"
#endif

#define max(i1,i2) (((i1) >= (i2)) ? (i1) : (i2))

extern int yylex(void);

#if YYBTYACC && !YYBYACC
void yyerror(const char *, ...);
#else
void yyerror(const char *);
#endif
%}

%token CONST

%start s

@attributes {short height,width,hanger; @void print; int val;
             short line,column; imageDesc imD;
	    } s e t f
@attributes {int val;} CONST

@macro binOpDefs(ee,t,label,op)
  @i @ee.1.line@ = @ee.0.line@ + 2;
  @i @ee.1.column@ = @ee.0.column@;
  @i @t.line@ = @ee.0.line@ + 2;
  @i @t.column@ = @ee.1.column@ + @ee.1.width@ + 3;
  @i @ee.0.height@ = max(@ee.1.height@,@t.height@) + 2;
  @i @ee.0.width@ = @ee.1.width@ + 3 + @t.width@;
  @i @ee.0.hanger@ = (@ee.1.hanger@ + @ee.1.width@ +3+ @t.hanger@) /2;
  @i @ee.val@ = @ee.1.val@ + @t.val@;
  @e ee.0.print: ee.0.line ee.0.column ee.0.hanger
	         ee.1.hanger ee.1.width t.hanger
		 ee.imD ee.1.print t.print
               ;
     {char *temp;
      temp = @ee.imD@.im + (@ee.imD@.w*@ee.0.line@)+@ee.0.column@;
      *(temp + @ee.0.hanger@) = label;
      temp += @ee.imD@.w;
      *(temp + (@ee.0.hanger@+@ee.1.hanger@)/2) = '/';
      *(temp + @ee.0.hanger@) = '|';
      *(temp+(@ee.0.hanger@+@ee.1.width@+3+@t.hanger@)/2) = '\\';
      *(temp + @ee.imD@.w + @ee.0.hanger@) = op;
     }
  @i @ee.1.imD@ = @ee.imD@;
  @i @t.imD@ = @ee.imD@;
@end

@macro dumRedDefs(t,f,label)
  @i @t.height@ = @f.height@ + 2;
  @i @t.width@ = @f.width@;
  @i @t.hanger@ = @f.hanger@;
  @i @f.line@ = @t.line@ + 2;
  @i @f.column@ = @t.column@;
  @i @t.val@ = @f.val@;
  @e t.print: t.line t.column t.hanger f.print t.imD;
     {char *temp;
      temp = @t.imD@.im+(@t.imD@.w*@t.line@)+@t.column@+@t.hanger@;
      *temp = label;
      *(temp+@t.imD@.w) = '|';
     }
  @i @f.imD@ = @t.imD@;
@end

%%
s	:
	@{  @e s.line:; {}
	    @e s.column:; {}
	    @e s.height:; {}
	    @e s.width:; {}
	    @e s.hanger:; {}
            @e s.imD:; {}
            @e s.print:; {}
            @e s.val:; {}
	@}
	|	s	e    	'\n' 	
	 @{ @decorate
            @i @s.line@ = 0;
	    @i @s.column@ = 0;
	    @i @s.height@ = @e.height@ + 2;
	    @i @s.width@ = @e.width@;
	    @i @s.hanger@ = @e.hanger@;
            @i @s.val@ = @e.val@;
	    @e s.imD: s.height s.width ; {
              int i;
              @s.imD@.h = @s.height@;
              @s.imD@.w = @s.width@ + 1;
	      if ((@s.imD@.im = (char *)malloc(@s.imD@.w * @s.imD@.h))
		  ==
		  ((char *)NULL)
		 )
		 {fprintf(stderr,"error in mallocing image storage\n");
	 	  exit(-1);
		 }
	      for (i=(@s.imD@.h*@s.imD@.w) - 1; i>=0;i--)
                (@s.imD@.im)[i] = ' ';
	      for (i=@s.imD@.h;i>0;i--)
                (@s.imD@.im)[i * @s.imD@.w - 1] = '\n';
              }
            @e s.print : e.print s.imD s.hanger s.val; {
	      (@s.imD@.im)[@s.hanger@] = 's';
	      (@s.imD@.im)[@s.imD@.w + @s.hanger@] = '|';
              showPT(@s.imD@);
              fprintf(stderr,"\n\n\n");
		       }
	    @i @e.line@ = @s.line@ + 2;
	    @i @e.column@ = @s.column@;
	    @i @e.imD@ = @s.imD@;
	@}
	;

e	:	e 	'+'	t{} 	
	 @{ binOpDefs(e,t,'e','+') @}

	|	t  {}
	 @{ dumRedDefs(e,t,'e') @}
	;

t	:	t	'*'	f{} 	
	 @{ binOpDefs(t,f,'t','*') @}

	| 	f	{} 	
	 @{ dumRedDefs(t,f,'t') @}
	;

f	:	'('	e	')' 	{}
         @{ @i @f.height@ = @e.height@ + 2;
	    @i @f.width@ = 1 + @e.width@ + 1;
	    @i @f.hanger@ = @e.hanger@ + 1;
	    @i @e.line@ = @f.line@ + 2;
            @i @e.column@ = @f.column@ + 1;
            @i @f.val@ = @e.val@;
            @e f.print: f.line f.column f.hanger f.width
		     e.line e.column e.hanger e.width
                     f.imD e.print
		   ;   {
              char *temp;
              temp = @f.imD@.im+(@f.imD@.w*@f.line@)+@f.column@+@f.hanger@;
              *temp = 'f';
              temp += @f.imD@.w;
              *temp = '|';
              *(temp-2) = '/';
              *(temp+2) = '\\';
              temp += @f.imD@.w;
              *(temp-3) = '(';
              *(temp+3) = ')';
			   }
	    @i @e.imD@ = @f.imD@;
	 @}

	|	CONST  	{}
	 @{ @i @f.height@ = 3;
	    @i @f.width@ = 7;
	    @i @f.hanger@ = 3;
            @i @f.val@ = @CONST.val@;
	    @e f.print : f.line f.column f.hanger
		      f.imD
		    ;  {
              char *temp;
	      temp = @f.imD@.im+(@f.imD@.w*@f.line@)+@f.column@+@f.hanger@;
              *temp = 'f';
              *(temp+@f.imD@.w) = '|';
              strncpy(temp+(2*@f.imD@.w)-2,"CONST",5);
			 }
	 @}
	;
%%

#if YYBTYACC && !YYBYACC
void yyerror(const char *s, ...)
{extern int yylineno;
 fprintf(stderr,"Near line %d: %s\n",yylineno,s);
 yyyabort();
}
#else
void yyerror(const char *s)
{fprintf(stderr,"%s\n",s);
 yyyabort();
}
#endif

void showPT(imageDesc image)
  {int i;
   for (i=0;i<(image.w * image.h);i++) fputc((image.im)[i],stderr);
   free(image.im);
  }


int yyparse(void);

int main()
  {yyparse();
  }
