/*
  tiny - C
*/

/* types of identifiers */
%token TK_IDENTIFIER

/* types of constants */
%token TK_ICONSTANT

/* simple operators */
%token TK_PTR TK_INC TK_DEC TK_LE TK_GE TK_EQ TK_NE
%token TK_AND TK_OR TK_DOT


/* keywords */
%token TK_EXTERN 
%token TK_INT 
%token TK_STRUCT
%token TK_IF TK_ELSE TK_WHILE TK_DO TK_FOR TK_RETURN TK_BREAK TK_SIZEOF

%{
enum symType  {INT_Type,BOOL_Type,VOID_Type,ERR_Type};
%}

/* This declares the attribute stack entry */
%union {
	double			d;
	int			i;
	char		       *s;
	struct typestruct      *t;
	unsigned int		u;
	struct sym             *y;
	struct node            *n;
	}

%type <i> TK_ICONSTANT 
%type <i> constant_expr

%start file		/* start symbol is file */

@autosyn type isStgRef isStg.Ref

@attributes {int val;} TK_ICONSTANT

@attributes {int val;
	     enum symType type;
	     bool isStgRef?, isStg.Ref, bad.Ref;
	    } constant_expr

@attributes {@autosyn enum symType type;
	     @autosyn bool isStgRef;
	    } primary_expr opte postfix_expr expr assignment_expr
	      unary_expr cast_expr
	      multiplicative_expr additive_expr shift_expr
	      relational_expr equality_expr and_expr exclusive_or_expr
	      inclusive_or_expr logical_and_expr logical_or_expr
	      conditional_expr

@macro constExpr(EXPR,SYM,TYPE??)
   @i @EXPR.type@ = TYPE;   
   @i @EXPR.val@ = @SYM.val@;   
   @i @EXPR.isStgRef@ = false;
@end

%%

file
	: external_definition
	| file external_definition
	;

external_definition
	: declaration			{ dclstate = globaldcl; }
	| function_definition		{ dclstate = globaldcl; }
	;

declaration
	/* This syntax is only permitted when declaring struct/union */
	: declaration_specifiers ';'
	| declaration_specifiers init_declarator_list ';'
	;

function_definition
	: function_head function_body
	;

function_head
	: declarator
		{ check_fstate(); }
	| declaration_specifiers declarator
		{ check_fstate(); }
	;

				/* returns a base type */
declaration_specifiers
	: ssc_specifier		/* this is only permitted when
				 *  dclstate == globaldcl
				 *  and ssc_specifier is extern
				 *  type defaults to int
				 */
		{ 
		  if (dclstate != globaldcl)
		    yyerror("illegal omission of type specifier");
		}
	| ssc_specifier type_specifier_list
	| type_specifier_list
	;

ssc_specifier
	: TK_EXTERN
	;

type_specifier_list
	: type_specifier
	;

type_specifier
	: TK_INT
	| struct_or_union_specifier
	;

/* struct_or_union_specifier should also accept struct_or_union '{' ... '}' */
struct_or_union_specifier
	: struct_def struct_declaration_list '}'
	| struct_or_union TK_IDENTIFIER
	;

struct_def
	: struct_or_union TK_IDENTIFIER '{'
	| struct_or_union '{'
	;

/* should permit unions too */
struct_or_union
	: TK_STRUCT
	;

struct_declaration_list
	: struct_declaration
	| struct_declaration_list struct_declaration
	;

struct_declaration
	: stype struct_declarator_list ';'
	;

stype
	: type_specifier_list
	;

struct_declarator_list
	: struct_declarator
	| struct_declarator_list ',' struct_declarator
	;

struct_declarator
	: sdeclarator
	;

sdeclarator
	: stars sdeclarator2
	;

stars
	: '*' stars
	|				/* null string */
	;

sdeclarator2
	: TK_IDENTIFIER
	| leftparen sdeclarator ')'
	| sdeclarator2 '[' constant_expr ']'
	;

leftparen
	: '('
	;

init_declarator_list
	: init_declarator
	| init_declarator_list ',' init_declarator      
		{
		  if (dclstate != fundcl)
		    check_dcl(varsdcl);
		}
	;

init_declarator
	: declarator
	;

declarator
	: declarator1
	;

declarator1
	: stars declarator2
	;

declarator2
	: TK_IDENTIFIER
	| leftparen declarator ')'
	| declarator2 '[' ']'
	| declarator2 '[' constant_expr ']'
	| declarator2 '(' ')'		
			{ check_dcl(posfundcl); }
	| declarator2 '(' parameter_declaration_list ')'
			{ check_dcl(fundcl); }
	;

parameter_declaration_list
	: identifier_list
	;

identifier_list
	: TK_IDENTIFIER
	| identifier_list ',' TK_IDENTIFIER
	;

function_body
	: param_declarations function_statement
	;

param_declarations
	:  /* nothing */		{ dclstate = funbody; }
	|
		{
		  dclstate = funbody;
		}
          declaration_list
	;

declaration_list
	: declaration
	| declaration_list declaration
	;

function_statement
	: '{' function_declarations statement_list '}'
	;

function_declarations
	: /* empty */
	| declaration_list
	;

statement_list
	:			/* empty string */
	| statement_list statement
	;

statement
	: compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	;

compound_statement
	: '{' statement_list '}'
	;

expression_statement
	: ';'
	| expr ';'
	;

selection_statement
	: TK_IF '(' expr ')' statement
	| TK_IF '(' expr ')' statement TK_ELSE statement
	;

iteration_statement
	: TK_WHILE '(' expr ')' statement
	| TK_DO statement TK_WHILE '(' expr ')' ';'
	| TK_FOR '(' opte ';' opte ';' opte ')' statement
	;

opte
	: /* nothing */
	  @{ @i @opte.type@     = BOOL_Type;
	     @i @opte.isStgRef@ = false;
	  @}
	| expr
	  @{ @i @opte.type@     = BOOL_Type;
	  @}
	;

jump_statement
	: TK_RETURN ';'
  	| TK_RETURN expr ';'
      	| TK_BREAK ';'
	;

primary_expr
	: TK_IDENTIFIER
	  		{
			  /* dump the symbol */
			  struct sym *vp;
			  extern struct sym *lookup();
			  extern void vdisplay(struct sym *);

			  if ((vp = lookup($<s>1)))
			    vdisplay(vp);
			  else
			    fprintf(stderr, "%s: not\
 defined\n", $<s>1);
			}
	  @{ @i @primary_expr.type@     = INT_Type;
	     @i @primary_expr.isStgRef@ = true;
	  @}
	| TK_ICONSTANT
	  @{ @i @primary_expr.type@     = INT_Type;
	     @i @primary_expr.isStgRef@ = false;
	  @}
	| '(' expr ')'
	  @{ @i @primary_expr.type@     = @expr.type@;
	     @i @primary_expr.isStgRef@ = false;
	  @}
	| primary_expr '[' expr ']'
	  @{ @i @primary_expr.0.type@     = @primary_expr.1.type@;
	     @i @primary_expr.0.isStgRef@ = @primary_expr.1.isStgRef@;
	  @}
	| primary_expr '(' ')'
	  @{ @i @primary_expr.0.type@     = @primary_expr.1.type@;
	     @i @primary_expr.0.isStgRef@ = false;
	  @}
	| primary_expr '(' argument_expr_list ')'
	  @{ @i @primary_expr.0.type@     = @primary_expr.1.type@;
	     @i @primary_expr.0.isStgRef@ = false;
	  @}
	| primary_expr '.' TK_IDENTIFIER
	  @{ @i @primary_expr.0.type@     = INT_Type;
	     @i @primary_expr.0.isStgRef@ = @primary_expr.1.isStgRef@;
	  @}
	| primary_expr TK_PTR TK_IDENTIFIER
	  @{ @i @primary_expr.0.type@     = INT_Type;
	     @i @primary_expr.0.isStgRef@ = @primary_expr.1.isStgRef@;
	  @}
	;

argument_expr_list
	: assignment_expr
	| argument_expr_list ',' assignment_expr
	;

postfix_expr
	: primary_expr
	| primary_expr TK_INC
	| primary_expr TK_DEC
	;

unary_expr
	: postfix_expr
	| TK_INC unary_expr
	| TK_DEC unary_expr
	| unary_operator cast_expr
	  @{ @i @unary_expr.isStgRef@ = false;
	  @}
	;

unary_operator
	: '&'
	| '*'
	| '+'
	| '-'
	| '!'
	;

cast_expr
	: unary_expr
	;

multiplicative_expr
	: cast_expr
	| multiplicative_expr '*' cast_expr
	  @{ @i @multiplicative_expr.0.type@     = INT_Type;
	     @i @multiplicative_expr.0.isStgRef@ = false;
	  @}
	| multiplicative_expr '/' cast_expr
	  @{ @i @multiplicative_expr.0.type@     = INT_Type;
	     @i @multiplicative_expr.0.isStgRef@ = false;
	  @}
	| multiplicative_expr '%' cast_expr
	  @{ @i @multiplicative_expr.0.type@     = INT_Type;
	     @i @multiplicative_expr.0.isStgRef@ = false;
	  @}
	;

additive_expr
	: multiplicative_expr
	| additive_expr '+' multiplicative_expr
	  @{ @i @additive_expr.0.type@     = INT_Type;
	     @i @additive_expr.0.isStgRef@ = false;
	  @}
	| additive_expr '-' multiplicative_expr
	  @{ @i @additive_expr.0.type@     = INT_Type;
	     @i @additive_expr.0.isStgRef@ = false;
	  @}
	;

shift_expr
	: additive_expr
	;

relational_expr
	: shift_expr
	  @{ @i @relational_expr.type@       = BOOL_Type;
	  @}
	| relational_expr '<' shift_expr
	  @{ @i @relational_expr.0.type@     = BOOL_Type;
	     @i @relational_expr.0.isStgRef@ = false;
	  @}
	| relational_expr '>' shift_expr
	  @{ @i @relational_expr.0.type@     = BOOL_Type;
	     @i @relational_expr.0.isStgRef@ = false;
	  @}
	| relational_expr TK_LE shift_expr
	  @{ @i @relational_expr.0.type@     = BOOL_Type;
	     @i @relational_expr.0.isStgRef@ = false;
	  @}
	| relational_expr TK_GE shift_expr
	  @{ @i @relational_expr.0.type@     = BOOL_Type;
	     @i @relational_expr.0.isStgRef@ = false;
	  @}
	;

equality_expr
	: relational_expr
	| equality_expr TK_EQ relational_expr
	  @{ @i @equality_expr.0.type@     = BOOL_Type;
	     @i @equality_expr.0.isStgRef@ = false;
	  @}
	| equality_expr TK_NE relational_expr
	  @{ @i @equality_expr.0.type@     = BOOL_Type;
	     @i @equality_expr.0.isStgRef@ = false;
	  @}
	;

and_expr
	: equality_expr
	;

exclusive_or_expr
	: and_expr
	;

inclusive_or_expr
	: exclusive_or_expr
	;

logical_and_expr
	: inclusive_or_expr
	;

logical_or_expr
	: logical_and_expr
	;

conditional_expr
	: logical_or_expr
	;

assignment_expr
	: conditional_expr
	| unary_expr assignment_operator assignment_expr
	  @{ @i @assignment_expr.type@     = @unary_expr.type@;
	     @i @assignment_expr.isStgRef@ = @unary_expr.isStgRef@;
	  @}
	;

/* should include all augmented assignment operators */
assignment_operator
	: '='
	;

expr
	: assignment_expr
	;

constant_expr
	: TK_ICONSTANT
	  @{ constExpr(constant_expr,TK_ICONSTANT,INT_Type) @}
	;

%%
